# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build a DeepLabV3Plus.

Reference:
  - [Encoder-Decoder with Atrous Separable Convolution for Semantic Image
      Segmentation](https://arxiv.org/pdf/1802.02611.pdf)
"""

import tensorflow as tf

from deeplab2 import common
from deeplab2.model import utils
from deeplab2.model.decoder import aspp
from deeplab2.model.layers import convolutions

layers = tf.keras.layers


class DeepLabV3Plus(tf.keras.layers.Layer):
    """A DeepLabV3+ decoder model.

    This model takes in low- and high-level features from an encoder and performs
    multi-scale context aggregation with the help of an ASPP layer on high-level
    features. These are concatenated with the low-level features and used as input
    to the classification head that is used to predict a semantic segmentation.
    """

    def __init__(self,
                 decoder_options,
                 deeplabv3plus_options,
                 bn_layer=tf.keras.layers.BatchNormalization):
        """Creates a DeepLabV3+ decoder of type tf.keras.layers.Layer.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          deeplabv3plus_options: Model options as defined in
            config_pb2.ModelOptions.DeeplabV3PlusOptions.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
        """
        super(DeepLabV3Plus, self).__init__(name='DeepLabv3Plus')

        self._high_level_feature_name = decoder_options.feature_key
        self._low_level_feature_name = deeplabv3plus_options.low_level.feature_key
        self._aspp = aspp.ASPP(decoder_options.aspp_channels,
                               decoder_options.atrous_rates,
                               bn_layer=bn_layer)

        # Layers for low-level feature transformation.
        self._project_conv_bn_act = convolutions.Conv2DSame(
            deeplabv3plus_options.low_level.channels_project,
            kernel_size=1,
            name='project_conv_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation='relu')

        # Layers for fusing low- and high-level features.
        self._fuse = convolutions.StackedConv2DSame(
            conv_type='depthwise_separable_conv',
            num_layers=2,
            output_channels=decoder_options.decoder_channels,
            kernel_size=3,
            name='fuse',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation='relu')

        self._final_conv = convolutions.Conv2DSame(
            deeplabv3plus_options.num_classes, kernel_size=1, name='final_conv')

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layer to global average pooling."""
        self._aspp.reset_pooling_layer()

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layer.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layer.
        """
        self._aspp.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._aspp.get_pool_size()

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_DEEPLABV3PLUS_ASPP: self._aspp,
            common.CKPT_DEEPLABV3PLUS_PROJECT_CONV_BN_ACT:
                self._project_conv_bn_act,
            common.CKPT_DEEPLABV3PLUS_FUSE: self._fuse,
            common.CKPT_SEMANTIC_LAST_LAYER: self._final_conv,
        }
        return items

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels]. Different keys should point to different features extracted
            by the encoder, e.g. low-level or high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the semantic prediction under key
          common.PRED_SEMANTIC_LOGITS_KEY.
        """
        low_level_features = features[self._low_level_feature_name]
        high_level_features = features[self._high_level_feature_name]

        high_level_features = self._aspp(high_level_features, training=training)

        low_level_features = self._project_conv_bn_act(low_level_features,
                                                       training=training)

        target_h = tf.shape(low_level_features)[1]
        target_w = tf.shape(low_level_features)[2]

        high_level_features = utils.resize_align_corners(
            high_level_features, [target_h, target_w])
        x = tf.concat([high_level_features, low_level_features], 3)
        x = self._fuse(x)

        return {common.PRED_SEMANTIC_LOGITS_KEY: self._final_conv(x)}



r"""
Uses the building detector to validate a construction event.

Given a site model:

* search for the K highest quality images at the start of the sequences.

* search for the K highest quality images at the end of the sequence.

* run the building detector on all of the chosen images.

* test if the site boundary intersects detections in the start images.

* test if the site boundary intersects detections in the end images.


.. code::

    StartTest | EndTest | Result
    ----------+---------+-------
        T     |    T    | Reject
    ----------+---------+-------
        T     |    F    | Reject
    ----------+---------+-------
        F     |    T    | Accept
    ----------+---------+-------
        F     |    F    | Reject
    ----------+---------+-------
        ?     |    T    | Accept
    ----------+---------+-------
        T     |    ?    | Reject
    ----------+---------+-------
        ?     |    F    | Reject
    ----------+---------+-------
        F     |    ?    | Accept
    ----------+---------+-------


Dataflow:

    * BAS outputs a region model with candidate site summaries

    * We should be given a kwcoco path that indexes all of the data we could
      look at. In MLOPs this will be a region cropped kwcoco path that indexes
      existant images on disk. In smartflow this will be a virtual kwcoco file
      that requires network access.
"""
#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class BuildingValidatorConfig(scfg.DataConfig):
    input_kwcoco = scfg.Value(None, help='input')

    input_region = scfg.Value(None)
    input_sites = scfg.Value(None)

    output_region_fpath = scfg.Value(None)
    output_sites_dpath = scfg.Value(None)
    output_site_manifest_fpath = scfg.Value(None)

    box_isect_threshold = scfg.Value(0.1, help='This fraction of the a detected building box must intersect the proposed polygon')
    box_score_threshold = scfg.Value(0.01, help='The detected building boxes must have a score higher than this')
    start_max_score = scfg.Value(1.0, help='The max building score needed in the start sequence')
    end_min_score = scfg.Value(0.1, help='The min building score needed in the end sequence')


IGNORE_CLASS_NAMES = ub.codeblock(
    '''
    Fixed-wing Aircraft
    Small Aircraft
    Cargo Plane
    Helicopter
    Passenger Vehicle
    Small Car
    Bus
    Pickup Truck
    Utility Truck
    Truck
    Cargo Truck
    Truck w/Box
    Truck Tractor
    Trailer
    Truck w/Flatbed
    Truck w/Liquid
    Crane Truck
    Railway Vehicle
    Passenger Car
    Cargo Car
    Flat Car
    Tank car
    Locomotive
    Maritime Vessel
    Motorboat
    Sailboat
    Tugboat
    Barge
    Fishing Vessel
    Ferry
    Yacht
    Container Ship
    Oil Tanker
    Engineering Vehicle
    Tower crane
    Container Crane
    Reach Stacker
    Straddle Carrier
    Mobile Crane
    Dump Truck
    Haul Truck
    Scraper/Tractor
    Front loader/Bulldozer
    Excavator
    Cement Mixer
    Ground Grader
    ''').split('\n')

DONT_IGNORE_CLASSNAMES = ub.codeblock(
    '''
    Aircraft Hangar
    Helipad
    Storage Tank
    Hut/Tent
    Shipping container lot
    Shipping Container
    Pylon
    Tower
    Shed
    Building
    Damaged Building
    Facility
    Construction Site
    Vehicle Lot
    ''').split('\n')


def main(cmdline=1, **kwargs):
    """
    Ignore:
        from geowatch.tasks.dino_detector.building_validator import *  # NOQA
        NODE_DPATH = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_fd298dba'
        from geowatch.utils.partial_format import fsubtemplate
        coco_fpath = fsubtemplate('$NODE_DPATH/pred_boxes.kwcoco.zip')
        from kwutil.util_path import coerce_patterned_paths
        site_summary = ub.Path(fsubtemplate('$NODE_DPATH/.pred/sv_crop/*/.pred/bas_poly/*/site_summaries_manifest.json'))
        site_summary = ub.Path(fsubtemplate('$NODE_DPATH/.pred/valicrop/*/.pred/bas_poly/*/site_summaries_manifest.json'))
        site_summary = coerce_patterned_paths(site_summary)[0]

        output_region_fpath = ub.Path(fsubtemplate('$NODE_DPATH/filtered_site_summaries.geojson'))

        kwargs = {
            'input_kwcoco': coco_fpath,
            'input_region': site_summary,
            'input_sites': (site_summary.parent / 'sites'),
            'output_region_fpath': output_region_fpath,
        }
        cmdline = 0

    Ignore:
        >>> import geowatch
        >>> dvc_data_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> coco_fpath = dvc_data_dpath / 'Drop6/imgonly-KR_R001.kwcoco.json'
        >>> region_fpath = dvc_data_dpath / 'annotations/drop6_hard_v1/region_models/KR_R001.geojson'

    Example:
        >>> # xdoctest: +SKIP
        >>> cmdline = 0
        >>> kwargs = dict()
        >>> main(cmdline=cmdline, **kwargs)
    """
    config = BuildingValidatorConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = ' + ub.urepr(config, nl=1))

    from geowatch.geoannots import geomodels
    from kwgis.utils import util_gis
    import kwcoco
    from geowatch.cli import reproject_annotations
    from kwcoco.util import util_json
    from geowatch.utils import process_context
    import os
    import safer
    import json

    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(config.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)
    filter_output = {
        'type': 'tracking_result',
        'info': [],
        'files': [],
    }
    # Track process info
    proc_context = process_context.ProcessContext(
        name='geowatch.tasks.dino_detector.building_validator', type='process',
        config=jsonified_config,
        track_emissions=False,
    )
    proc_context.start()
    filter_output['info'].append(proc_context.obj)

    input_coco = kwcoco.CocoDataset(config.input_kwcoco)
    region_model = geomodels.RegionModel.coerce(config.input_region)
    output_region_fpath = ub.Path(config.output_region_fpath)

    input_site_fpaths = util_gis.coerce_geojson_paths(config.input_sites)
    # set(input_coco.annots().lookup('track_id', None))

    site_to_site_fpath = ub.udict({
        p.stem: p for p in input_site_fpaths
    })
    input_sites = list(geomodels.SiteModel.coerce_multiple(site_to_site_fpath.values()))

    site_id_to_summary = ub.udict()
    for summary in region_model.site_summaries():
        assert summary.site_id not in site_id_to_summary
        site_id_to_summary[summary.site_id] = summary
    ##

    # Ensure caches
    for summary in site_id_to_summary.values():
        summary['properties'].setdefault('cache', {})
    for site in input_sites:
        site.header['properties'].setdefault('cache', {})

    output_kwcoco = reproject_annotations.main(
        cmdline=0, src=input_coco.copy(),
        dst='return',
        region_models=config.input_region,
        status_to_catname={'system_confirmed': 'positive'},
        role='pred_poly',
        validate_checks=False,
        clear_existing=False,
    )

    if 0:
        from geowatch.utils import util_framework
        region_models = [region_model]
        site_models = input_sites
        util_framework.check_region_and_site_models_agree(region_models, site_models)

    site_to_decisions = make_decisions(output_kwcoco, site_id_to_summary, config)

    num_accept = sum(d['accept'] for s, d in site_to_decisions.items())
    print(f'Filter to {num_accept} / {len(site_id_to_summary)} sites')

    # Enrich each site summary with the decision reason and update status
    for site_id, decision in site_to_decisions.items():
        sitesum = site_id_to_summary[site_id]

        # Change the status of sites to "system_rejected" instead of droping them
        if not decision['accept']:
            sitesum['properties']['status'] = 'system_rejected'

        if 'cache' not in sitesum['properties']:
            sitesum['properties']['cache'] = {}

        sitesum['properties']['cache']['dino_decision'] = decision

    # Copy the site models and update their header with new summary
    # information.
    output_sites_dpath = ub.Path(config.output_sites_dpath)
    output_sites_dpath.ensuredir()
    out_site_fpaths = []

    for old_site in input_sites:
        site_id = old_site.site_id
        old_fpath = site_to_site_fpath[site_id]
        new_fpath = output_sites_dpath / old_fpath.name
        new_summary = site_id_to_summary[site_id]
        old_site.header['properties']['status'] = new_summary['properties']['status']
        if 'cache' not in old_site.header['properties']:
            old_site.header['properties']['cache'] = {}
        old_site.header['properties']['cache'].update(new_summary['properties']['cache'])
        new_fpath.write_text(old_site.dumps())
        out_site_fpaths.append(new_fpath)

    # Write the updated site summaries in a new region model
    new_summaries = list(site_id_to_summary.values())
    new_region_model = geomodels.RegionModel.from_features(
        [region_model.header] + list(new_summaries))
    output_region_fpath.parent.ensuredir()

    if 1:
        from geowatch.utils import util_framework
        region_models = [region_model]
        site_models = input_sites
        util_framework.check_region_and_site_models_agree([new_region_model], site_models)

    print(f'Write filtered region model to: {output_region_fpath}')
    with safer.open(output_region_fpath, 'w', temp_file=not ub.WIN32) as file:
        json.dump(new_region_model, file, indent=4)

    # from kwutil.util_json import debug_json_unserializable
    # debug_json_unserializable(new_region_model)

    proc_context.stop()

    if config.output_site_manifest_fpath is not None:
        filter_output['files'] = [os.fspath(p) for p in out_site_fpaths]
        print(f'Write filtered site result to {config.output_site_manifest_fpath}')
        with safer.open(config.output_site_manifest_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(filter_output, file, indent=4)


def make_decisions(output_kwcoco, site_id_to_summary, config):
    """
    Make yes/no decision about sites with dino predictions in the kwcoco file.

    Returns:
        Dict: mapping from site-ids to the decision made for it
    """
    from kwcoco.util import util_json
    from kwutil import util_time
    # Enrich all sites with features (evidence) and decisions.
    site_to_decisions = {
        s: {
            'type': 'dino_decision',
            'accept': True,
            'why': None,
            'features': None,
        }
        for s in site_id_to_summary.keys()
    }
    for video_id in ub.ProgIter(output_kwcoco.videos(), desc='validate sites'):
        video = output_kwcoco.index.videos[video_id]
        video_name = video['name']  # the vide name should be the site id
        summary = site_id_to_summary[video_name]
        site_id = summary.site_id
        video_images = output_kwcoco.images(video_id=video_id)

        start_date = summary.start_date
        end_date = summary.end_date

        # Get the starting and ending observations
        start_images = []
        end_images = []
        for coco_img in video_images.coco_images:
            img_time = util_time.coerce_datetime(coco_img['date_captured'])
            dist_start = abs(img_time - start_date)
            dist_end = abs(img_time - end_date)
            if dist_end < dist_start:
                end_images.append(coco_img)
            else:
                start_images.append(coco_img)

        start_features = []
        end_features = []
        try:
            for coco_img in start_images:
                feat = building_in_image_features(coco_img, site_id, config)
                feat['datetime'] = coco_img['date_captured']
                feat['type'] = 'start_feature'
                start_features.append(feat)

            for coco_img in end_images:
                feat = building_in_image_features(coco_img, site_id, config)
                feat['datetime'] = coco_img['date_captured']
                feat['type'] = 'end_feature'
                end_features.append(feat)

            if len(start_features) and len(end_features):
                max_start_score = max(f['max_score'] for f in start_features)
                max_end_score = max(f['max_score'] for f in end_features)
                accept = bool(
                    max_start_score <= config.start_max_score and
                    max_end_score >= config.end_min_score
                )
                why = 'threshold'
            else:
                # Unobservable case, automatically accept
                accept = True
                why = 'unobservable'
        except CouldNotValidate:
            accept = True
            why = 'CouldNotValidate'
        decision = site_to_decisions[site_id]
        decision['accept'] = accept
        decision['why'] = why
        decision['features'] = start_features + end_features
        decision = util_json.ensure_json_serializable(decision)
        site_to_decisions[site_id] = decision
    return site_to_decisions


def building_in_image_features(coco_img, site_id, config):
    import numpy as np
    import kwimage
    import geopandas as gpd
    import warnings
    annots = coco_img.annots()
    flags = np.array([r == 'pred_poly' for r in annots.lookup('role', None)], dtype=bool)
    box_annots = annots.compress(~flags)
    poly_annots = annots.compress(flags)
    is_main_poly = [t == site_id for t in poly_annots.lookup('track_id')]
    main_poly_annots = poly_annots.compress(is_main_poly)
    if len(main_poly_annots) > 1:
        warnings.warn('FIXME: Len of "main-poly-annots" is not 1.')
    elif len(main_poly_annots) == 0:
        raise CouldNotValidate('We dont expect to be here')

    flags = [cname not in IGNORE_CLASS_NAMES for cname in box_annots.category_names]
    box_annots = box_annots.compress(flags)

    iooa_thresh = config.box_isect_threshold
    score_thresh = config.box_score_threshold

    box_gdf = gpd.GeoDataFrame({
        'geometry': box_annots.boxes.to_shapely(),
        'class': box_annots.category_names,
        'score': box_annots.lookup('score'),
    })
    box_gdf = box_gdf[box_gdf.score > score_thresh]

    proposal_gdf = gpd.GeoDataFrame({
        'geometry': [
            kwimage.MultiPolygon.coerce(p).to_shapely()
            for p in main_poly_annots.lookup('segmentation')
        ],
    })
    proposal_geom = proposal_gdf.iloc[0]['geometry']

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value encountered in intersection')
        intersections = box_gdf.intersection(proposal_geom)
        bad_flags = (intersections.isnull() | intersections.is_empty)
        isect_flags = (~bad_flags) & (intersections.is_valid)
        feat = {
            'max_iou': 0,
            'max_iooa': 0,
            'max_score': 0,
            'num_cands': 0,
        }
        if isect_flags.any():
            cand_boxes = box_gdf[isect_flags]
            cand_isects = intersections[isect_flags]
            cand_iooa1 = cand_isects.area / cand_boxes.area
            ioaa_flags = cand_iooa1 > iooa_thresh
            if ioaa_flags.any():
                cand_boxes = cand_boxes[ioaa_flags]

                cand_unions = cand_boxes.union(proposal_geom)
                cand_ious = cand_isects.area / cand_unions.area

                feat['max_iou'] = cand_ious.max()
                feat['max_iooa'] = cand_iooa1.max()
                feat['max_score'] = cand_boxes['score'].max()
                feat['num_cands'] = len(cand_boxes)
        return feat


class CouldNotValidate(Exception):
    ...


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/geowatch/tasks/dino_detector/building_validator.py
        python -m geowatch.tasks.dino_detector.building_validator
    """
    main()

"""

Ignore:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    geowatch schedule --params="
        matrix:
            bas_pxl.package_fpath:
                - $DVC_EXPT_DPATH/models/fusion/Drop6-MeanYear10GSD/packages/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2_epoch3_step941.pt
            bas_pxl.test_dataset:
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R002.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-BR_R002.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-CH_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-NZ_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip
            bas_pxl.chip_overlap: 0.3
            bas_pxl.chip_dims:
                - auto
            bas_pxl.time_span:
                - auto
            bas_pxl.time_sampling:
                # - auto
                # - soft5
                - soft4
            bas_poly.thresh:
                - 0.33
                #- 0.38
                #- 0.4
            bas_poly.inner_window_size:
                - 1y
                #- null
            bas_poly.inner_agg_fn:
                - mean
            bas_poly.norm_ord:
                #- 1
                - inf
            bas_poly.polygon_simplify_tolerance:
                - 1
            bas_poly.agg_fn:
                - probs
            bas_poly.resolution:
                - 10GSD
            bas_poly.moving_window_size:
                - null
                #- 1
            bas_poly.poly_merge_method:
                - 'v2'
                #- 'v1'
            bas_poly.min_area_square_meters:
                - 7200
            bas_poly.max_area_square_meters:
                - 8000000
            bas_poly.boundary_region: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_poly_eval.true_site_dpath: $DVC_DATA_DPATH/annotations/drop6/site_models
            bas_poly_eval.true_region_dpath: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_pxl.enabled: 1
            bas_pxl_eval.enabled: 1
            bas_poly.enabled: 1
            bas_poly_eval.enabled: 1
            bas_poly_viz.enabled: 0
            sv_crop.enabled: 1
            sv_crop.minimum_size: "256x256@2GSD"
            sv_crop.num_start_frames: 3
            sv_crop.num_end_frames: 3
            sv_crop.context_factor: 1.5
            sv_dino_boxes.enabled: 1
            sv_dino_boxes.package_fpath: $DVC_EXPT_DPATH/models/kitware/xview_dino.pt
            sv_dino_boxes.window_dims:
                - 256
                - 512
                - 768
                - 1024
                # - 1536
            sv_dino_boxes.window_overlap:
                - "0.5"
            sv_dino_boxes.fixed_resolution:
                - "1GSD"
                - "2GSD"
                - "2.5GSD"
                - "3GSD"
                - "3.3GSD"
            sv_dino_filter.box_isect_threshold:
                - 0.1
            sv_dino_filter.box_score_threshold:
                - 0.01
            sv_dino_filter.start_max_score:
                - 1.0
                - 0.8
                # - 0.5
            sv_dino_filter.end_min_score:
                - 0.0
                # - 0.05
                - 0.1
                - 0.15
                - 0.2
                - 0.25
                - 0.3
                # - 0.4
                - 0.5
        submatrices:
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R001.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R002.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-BR_R002.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-BR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-CH_R001.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-CH_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-NZ_R001.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-NZ_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-AE_R001.kwcoco.json
        " \
        --root_dpath="$DVC_EXPT_DPATH/_mlops_eval10_baseline" \
        --devices="0," --tmux_workers=8 \
        --backend=tmux --queue_name "_mlops_eval10_baseline" \
        --pipeline=bas_building_vali --skip_existing=1 \
        --run=1

DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
gwmlops aggregate \
    --pipeline=bas_building_vali \
    --target \
        "$DVC_EXPT_DPATH/_mlops_eval10_baseline" \
    --resource_report=0 \
    --rois='[KR_R002,KR_R001,BR_R002,CH_R001,NZ_R001,AE_R001]' \
    --stdout_report="
        top_k: 5
        per_group: 1
        macro_analysis: 0
        analyze: 0
        reference_region: final
        shorten: 1
    " --eval_nodes="[sv_poly_eval]" \
    --plot_params="
        enabled: True
    "

    --rois='[NZ_R001,KR_R001]' \



        print_models: True

geowatch align \
    --src "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6/imgonly-KR_R001.kwcoco.json" \
    --dst "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/sv_crop/valicrop_id_2e8c8dc3/sv_crop.kwcoco.zip" \
    --regions="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_dc32b2a6/site_summaries_manifest.json" \
    --site_summary=True \
    --verbose="1" \
    --workers="32" \
    --aux_workers="4" \
    --debug_valid_regions="False" \
    --visualize="False" \
    --keep="img" \
    --geo_preprop="auto"  \
    --minimum_size="256x256@2GSD" \
    --num_start_frames="3" \
    --num_end_frames="3" \
    --context_factor="1.5" \
    --include_sensors="WV" \
    --force_nodata="-9999" \
    --rpc_align_method="orthorectify" \
    --target_gsd="2" \
    --force_min_gsd="2" \
    --convexify_regions="True"

geowatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/sv_crop/valicrop_id_2e8c8dc3/sv_crop.kwcoco.zip --smart

python ~/code/watch/dev/wip/grid_sitevali_crops.py /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/sv_crop/valicrop_id_2e8c8dc3/_viz_*


python -m geowatch.tasks.dino_detector.predict \
    --package_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/kitware/xview_dino.pt" \
    --coco_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/sv_crop/valicrop_id_2e8c8dc3/sv_crop.kwcoco.zip" \
    --out_coco_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/pred_boxes.kwcoco.zip" \
    --device="0" \
    --data_workers="2" \
    --fixed_resolution="1.0GSD" \
    --window_dims="2048" \
    --batch_size="1"

geowatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/pred_and_truth.kwcoco.zip \
    --resolution=2GSD \
    --smart \
    --ann_score_thresh=0.3 \
    --viz_dpath /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/_vizme

python ~/code/watch/dev/wip/grid_sitevali_crops.py --sub=_anns \
    /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/_vizme


NODE_DPATH=/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_fd298dba
DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)

geowatch reproject \
        --src "$NODE_DPATH/pred_boxes.kwcoco.zip" \
        --dst "$NODE_DPATH/pred_boxes_with_polys.kwcoco.zip" \
        --region_models "$NODE_DPATH/.pred/sv_crop/*/.pred/bas_poly/*/site_summaries_manifest.json" \
        --status_to_catname="{system_confirmed: positive}" \
        --role=pred_poly \
        --validate_checks=False \
        --clear_existing=False

geowatch reproject \
        --src "$NODE_DPATH/pred_boxes_with_polys.kwcoco.zip" \
        --dst "$NODE_DPATH/pred_and_truth.kwcoco.zip" \
        --region_models="$DVC_DATA_DPATH/annotations/drop6/region_models/*.geojson" \
        --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/*.geojson" \
        --status_to_catname="{system_confirmed: positive}" \
        --role=truth \
        --clear_existing=False

gw visualize --smart 1 \
    --ann_score_thresh 0.5 \
    --draw_labels False \
    --alpha 0.5 \
    --src $NODE_DPATH/pred_and_truth.kwcoco.zip \
    --viz_dpath $NODE_DPATH/_vizme \

python ~/code/watch/dev/wip/grid_sitevali_crops.py \
    --sub=_anns \
    $NODE_DPATH/_vizme

python -m geowatch.tasks.dino_detector.building_validator \
    --input_kwcoco "$NODE_DPATH/pred_boxes.kwcoco.zip" \
    --input_region $NODE_DPATH/.pred/sv_crop/*/.pred/bas_poly/*/site_summaries_manifest.json \
    --output_region_fpath "$NODE_DPATH/filtered_summaries.json" \
    --box_isect_threshold 0.1 \
    --box_score_threshold 0.1 \
    --start_max_score 0.0 \
    --end_min_score 0.3

python -m geowatch.tasks.dino_detector.building_validator \
    --input_kwcoco="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_663bd461/pred_boxes.kwcoco.zip" \
    --input_region="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_f8061df6/site_summaries_manifest.json" \
    --input_sites="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_f8061df6/sites_manifest.json" \
    --output_region_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_region.geojson" \
    --output_sites_dpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_sites" \
    --output_site_manifest_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_site_manifest.json" \
    --box_isect_threshold="0.1" \
    --box_score_threshold="0.1" \
    --start_max_score="0.1" \
    --end_min_score="0.3"


"""



import math
# import sys
# import time

import skimage.morphology
import skimage.io
from PIL import Image, ImageDraw
import numpy as np
import shapely.geometry
import shapely.affinity
# from lydorn_utils import print_utils
# from scipy.ndimage.morphology import distance_transform_edt
import cv2 as cv

from functools import partial

import torch_lydorn.torchvision


class Rasterize(object):
    """Rasterize polygons"""

    def __init__(self, fill=True, edges=True, vertices=True, line_width=3, antialiasing=False, return_distances=False,
                 return_sizes=False):
        self.fill = fill
        self.edges = edges
        self.vertices = vertices
        self.line_width = line_width
        self.antialiasing = antialiasing

        if not return_distances and not return_sizes:
            self.raster_func = partial(draw_polygons, fill=self.fill, edges=self.edges, vertices=self.vertices,
                                       line_width=self.line_width, antialiasing=self.antialiasing)
        elif return_distances and return_sizes:
            self.raster_func = partial(compute_raster_distances_sizes, fill=self.fill, edges=self.edges, vertices=self.vertices,
                                       line_width=self.line_width, antialiasing=self.antialiasing)
        else:
            raise NotImplementedError

    def __call__(self, image, polygons):
        """
        If distances is True, also returns distances image
        (sum of distance to closest and second-closest annotation for each pixel).
        Same for sizes (size of annotation the pixel belongs to).

        """
        size = (image.shape[0], image.shape[1])
        out = self.raster_func(polygons, size)
        return out


def compute_raster_distances_sizes(polygons, shape, fill=True, edges=True, vertices=True, line_width=3, antialiasing=False):
    """
    Returns:
         - distances: sum of distance to closest and second-closest annotation for each pixel.
         - size_weights: relative size (normalized by image area) of annotation the pixel belongs to.
    """
    assert isinstance(polygons, list), "polygons should be a list"

    # Filter out zero-area polygons
    polygons = [polygon for polygon in polygons if 0 < polygon.area]

    # tic = time.time()

    channel_count = fill + edges + vertices
    polygons_raster = np.zeros((*shape, channel_count), dtype=np.uint8)
    distance_maps = np.ones((*shape, len(polygons)))  # Init with max value (distances are normed)
    sizes = np.ones(shape)  # Init with max value (sizes are normed)
    image_area = shape[0] * shape[1]
    for i, polygon in enumerate(polygons):
        minx, miny, maxx, maxy = polygon.bounds
        mini = max(0, math.floor(miny) - 2 * line_width)
        minj = max(0, math.floor(minx) - 2 * line_width)
        maxi = min(polygons_raster.shape[0], math.ceil(maxy) + 2 * line_width)
        maxj = min(polygons_raster.shape[1], math.ceil(maxx) + 2 * line_width)
        bbox_shape = (maxi - mini, maxj - minj)
        bbox_polygon = shapely.affinity.translate(polygon, xoff=-minj, yoff=-mini)
        bbox_raster = draw_polygons([bbox_polygon], bbox_shape, fill, edges, vertices, line_width, antialiasing)
        polygons_raster[mini:maxi, minj:maxj] = np.maximum(polygons_raster[mini:maxi, minj:maxj], bbox_raster)
        bbox_mask = 0 < np.sum(bbox_raster, axis=2)  # Polygon interior + edge + vertex
        if bbox_mask.max():  # Make sure mask is not empty
            polygon_mask = np.zeros(shape, dtype=np.bool)
            polygon_mask[mini:maxi, minj:maxj] = bbox_mask
            polygon_dist = cv.distanceTransform(1 - polygon_mask.astype(np.uint8), distanceType=cv.DIST_L2, maskSize=cv.DIST_MASK_5,
                                                dstType=cv.CV_64F)
            polygon_dist /= (polygon_mask.shape[0] + polygon_mask.shape[1])  # Normalize dist
            distance_maps[:, :, i] = polygon_dist

            selem = skimage.morphology.disk(line_width)
            bbox_dilated_mask = skimage.morphology.binary_dilation(bbox_mask, selem=selem)
            sizes[mini:maxi, minj:maxj][bbox_dilated_mask] = polygon.area / image_area

    polygons_raster = np.clip(polygons_raster, 0, 255)
    # skimage.io.imsave("polygons_raster.png", polygons_raster)

    if edges:
        edge_channels = -1 + fill + edges
        # Remove border edges because they correspond to cut buildings:
        polygons_raster[:line_width, :, edge_channels] = 0
        polygons_raster[-line_width:, :, edge_channels] = 0
        polygons_raster[:, :line_width, edge_channels] = 0
        polygons_raster[:, -line_width:, edge_channels] = 0

    distances = compute_distances(distance_maps)
    # skimage.io.imsave("distances.png", distances)

    distances = distances.astype(np.float16)
    sizes = sizes.astype(np.float16)

    # toc = time.time()
    # print(f"Rasterize {len(polygons)} polygons: {toc - tic}s")

    return polygons_raster, distances, sizes


def compute_distances(distance_maps):
    distance_maps.sort(axis=2)
    distance_maps = distance_maps[:, :, :2]
    distances = np.sum(distance_maps, axis=2)
    return distances


def draw_polygons(polygons, shape, fill=True, edges=True, vertices=True, line_width=3, antialiasing=False):
    assert isinstance(polygons, list), "polygons should be a list"
    # print('LEN of polygon is :', len(polygons), '\n')

    assert isinstance(polygons[0], shapely.geometry.Polygon), "polygon should be a shapely.geometry.Polygon"

    if antialiasing:
        draw_shape = (2 * shape[0], 2 * shape[1])
        polygons = [shapely.affinity.scale(polygon, xfact=2.0, yfact=2.0, origin=(0, 0)) for polygon in polygons]
        line_width *= 2
    else:
        draw_shape = shape
    # Channels
    fill_channel_index = 0  # Always first channel
    edges_channel_index = fill  # If fill == True, take second channel. If not then take first
    vertices_channel_index = fill + edges  # Same principle as above
    channel_count = fill + edges + vertices
    im_draw_list = []
    for channel_index in range(channel_count):
        im = Image.new("L", (draw_shape[1], draw_shape[0]))
        im_px_access = im.load()  # NOQA
        draw = ImageDraw.Draw(im)
        im_draw_list.append((im, draw))

    for polygon in polygons:
        if fill:
            draw = im_draw_list[fill_channel_index][1]
            draw.polygon(polygon.exterior.coords, fill=255)
            for interior in polygon.interiors:
                draw.polygon(interior.coords, fill=0)
        if edges:
            draw = im_draw_list[edges_channel_index][1]
            draw.line(polygon.exterior.coords, fill=255, width=line_width)
            for interior in polygon.interiors:
                draw.line(interior.coords, fill=255, width=line_width)
        if vertices:
            draw = im_draw_list[vertices_channel_index][1]
            for vertex in polygon.exterior.coords:
                torch_lydorn.torchvision.transforms.functional.draw_circle(draw, vertex, line_width / 2, fill=255)
            for interior in polygon.interiors:
                for vertex in interior.coords:
                    torch_lydorn.torchvision.transforms.functional.draw_circle(draw, vertex, line_width / 2, fill=255)

    im_list = []
    if antialiasing:
        # resize images:
        for im_draw in im_draw_list:
            resize_shape = (shape[1], shape[0])
            im_list.append(im_draw[0].resize(resize_shape, Image.BILINEAR))
    else:
        for im_draw in im_draw_list:
            im_list.append(im_draw[0])

    # Convert image to numpy array with the right number of channels
    array_list = [np.array(im) for im in im_list]
    array = np.stack(array_list, axis=-1)
    return array


def _rasterize_coco(image, polygons):
    import pycocotools.mask as cocomask

    image_size = image.shape[:2]
    mask = np.zeros(image_size)
    for polygon in polygons:
        rle = cocomask.frPyObjects([np.array(polygon.exterior.coords).reshape(-1)], image_size[0], image_size[1])
        m = cocomask.decode(rle)

        for i in range(m.shape[-1]):
            mi = m[:, :, i]
            mi = mi.reshape(image_size)
            mask += mi
    return mask


def _test():
    import skimage.io

    rasterize = Rasterize(fill=True, edges=False, vertices=False, line_width=2, antialiasing=True, return_distances=True, return_sizes=True)

    image = np.zeros((300, 300))
    polygons = [
        shapely.geometry.Polygon([
            [10.5, 10.5],
            [100, 10],
            [100, 150],
            [10, 100],
            [10, 10],
        ]),
        shapely.geometry.Polygon([
            [10 + 150, 10],
            [100 + 150, 10],
            [100 + 150, 100],
            [10 + 150, 100],
            [10 + 150, 10],
        ]),
    ]
    polygons_raster, distances, size_weights = rasterize(image, polygons)

    skimage.io.imsave('rasterize.polygons_raster.png', polygons_raster)
    skimage.io.imsave('rasterize.distances.png', distances)
    skimage.io.imsave('rasterize.size_weights.png', size_weights)

    # Rasterize with pycocotools
    coco_mask = _rasterize_coco(image, polygons)
    skimage.io.imsave('rasterize.coco_mask.png', coco_mask)


if __name__ == "__main__":
    _test()



#!/usr/bin/env python3
import scriptconfig as scfg


class FindRecentCheckpointCLI(scfg.DataConfig):
    """
    Helper script to lookup the most recent checkpoint.

    Not sure what the best home for this script is. Useful to help make train
    scripts more consise.  Perhaps this is part of the mlops CLI?

    Usage
    -----

    This prints out extra arguments to be used at the end of a lightning CLI
    invocation.  As such, you should ensure its contents are read into a bash
    array, and that array should be passed to the invocation such that any
    bash-level word splitting is explicit.

    In short this is the following pattern that should be used

    .. code:: bash

        PREV_CHECKPOINT_TEXT=$(python -m geowatch.cli.experimental.find_recent_checkpoint --default_root_dir="$DEFAULT_ROOT_DIR")
        echo "PREV_CHECKPOINT_TEXT = $PREV_CHECKPOINT_TEXT"
        if [[ "$PREV_CHECKPOINT_TEXT" == "None" ]]; then
            PREV_CHECKPOINT_ARGS=()
        else
            PREV_CHECKPOINT_ARGS=(--ckpt_path "$PREV_CHECKPOINT_TEXT")
        fi
        echo "${PREV_CHECKPOINT_ARGS[@]}"


    This method of usage will do nothing when there is no checkpoint, and add
    the appropriate restart argument when something is needed.
    """
    default_root_dir = scfg.Value(None, help='the default root dir passed to lightning', position=1)
    allow_last = scfg.Value(True, isflag=True, help='if True, then prevent the last.ckpt from being chosen')
    as_cli_arg = scfg.Value(False, isflag=True, help='if True, print text that can be used to extend a lightning CLI invocation')

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> import ubelt as ub
            >>> from geowatch.cli.experimental.find_recent_checkpoint import *  # NOQA
            >>> cmdline = 0
            >>> # Make a dummy train directory
            >>> default_root_dir = ub.Path.appdir('geowatch/tests/find_recent_checkpoint')
            >>> fake_dpath = (default_root_dir / 'lightning_logs/version_0/checkpoints').ensuredir()
            >>> (fake_dpath / 'pretend.ckpt').write_text('dummy')
            >>> kwargs = dict(default_root_dir=default_root_dir)
            >>> cls = FindRecentCheckpointCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
            >>> kwargs['as_cli_arg'] = True
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import ubelt as ub
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        root_dir = ub.Path(config.default_root_dir)
        checkpoints = list((root_dir / 'lightning_logs').glob('version_*/checkpoints/*.ckpt'))
        if len(checkpoints) == 0:
            print('None')
        else:
            version_to_checkpoints = ub.group_items(checkpoints, key=lambda x: int(x.parent.parent.name.split('_')[-1]))
            max_version = max(version_to_checkpoints)
            candidates = version_to_checkpoints[max_version]
            if not config.allow_last:
                checkpoints = [p for p in checkpoints if 'last.ckpt' not in p.name]
            checkpoints = sorted(candidates, key=lambda p: p.stat().st_mtime)
            chosen = checkpoints[-1]
            print(chosen)

__cli__ = FindRecentCheckpointCLI
main = __cli__.main

if __name__ == '__main__':
    """
    CommandLine:
        python -m geowatch.cli.experimental.find_recent_checkpoint
    """
    main()



import portion  # NOQA
import ubelt as ub


class PolygonExtractor:
    """
    Given a timesequence of heatmaps, extract spatially static polygons.

    This class is being developed on ``dev/refine-ac-polys`` in the file:
    geowatch/tasks/tracking/polygon_extraction.py

    Distributed Tweaking on:
        https://colab.research.google.com/drive/1NEJpm36LviesZb45qy59myezi7JHu0bI#scrollTo=G8kHgCXSI3VS

    Example:
        >>> from geowatch.tasks.tracking.polygon_extraction import *  # NOQA
        >>> cls = PolygonExtractor
        >>> self = PolygonExtractor.demo(real_categories=True)
        >>> self.config['algo'] = 'crall'
        >>> print(f'self.heatmap_thwc.shape={self.heatmap_thwc.shape}')

        >>> polys, info = self.predict_polygons(return_info=True)
        >>> label_mask = info['label_mask']

        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> from geowatch.utils import util_kwimage
        >>> kwplot.autompl()
        >>> stacked = self.draw_timesequence()
        >>> canvas = label_mask.colorize()
        >>> for p in polys:
        >>>     canvas = p.draw_on(canvas, fill=0, edgecolor='black')
        >>> pnum_ = kwplot.PlotNums(nSubplots=2)
        >>> kwplot.imshow(stacked, pnum=pnum_(), title='Colorized Time Series')
        >>> kwplot.imshow(canvas, pnum=pnum_(), title='Cluster Labels')
        >>> kwplot.show_if_requested()
    """

    def __init__(self, heatmap_thwc, heatmap_time_intervals=None, bounds=None,
                 classes=None, config=None):
        """
        Args:

            heatmap_thwc (ndarray): An heatmap with dimensions [T, H, W, C]
                where T is number of timesteps and C are feature map reponces
                for each channel.

            heatmap_time_intervals (None | List[Tuple[datetime, datetime]]):
                A list of length T rows. Each row corresponds to a timestep
                in the heatmaps, and contains a tuple of two datetimes:
                corresponding to the start and end of times that
                information corresponds to. If a heatmap only corrseponds to a
                single time, then start and end will be the same.

            bounds (kwimage.MultiPolygon | None):
                if given, this is the polygon detected at BAS time.

            classes (None | List[str]):
                If specified, this corresponds to the class in each position
                of the C dimension.

            config (None | Dict):
                optional configuration

        """
        self.heatmap_thwc = heatmap_thwc
        self.bounds = bounds
        self.classes = classes
        self.heatmap_time_intervals = heatmap_time_intervals

        # For recording intermediate parts of the process.
        self._intermediates = {}

        self.default_config = {
            # 'algo': 'meanshift',
            'algo': 'kmeans',
            'workers': 'avail',
            'robust_normalize': False,
            'positional_encoding': True,
            'positional_encoding_scale': 1.0,
            'polygon_simplify_tolerance': None,
            'scale_factor': 'auto',
            'thresh': 0.3,
            'viz_out_dir': None,
        }
        if config is None:
            config = {}
        self.config = self.default_config | config

    def _predict_leotta(self):
        import numpy as np
        from scipy import ndimage
        heatmap_thwc = self.heatmap_thwc
        assert self.classes is not None

        # import kwarray
        # heatmap_thwc = kwarray.robust_normalize(heatmap_thwc)

        _t, _h, _w, _c = self.heatmap_thwc.shape
        if self.bounds is not None:
            mask = self.bounds.to_mask(dims=(_h, _w)).data

        salient_idx = self.classes.index('ac_salient')
        active_index = self.classes.index('Active Construction')
        # active_index = 2

        # commbine active and saliency and remove NANs
        active_vol = heatmap_thwc[:, :, :, active_index]
        saliency_vol = heatmap_thwc[:, :, :, salient_idx]
        vol = impute_nans2(np.multiply(active_vol, saliency_vol)) * mask[None, :, :]

        # morphological clean-up, a little more in the time direction
        vol = ndimage.grey_closing(vol, (7, 3, 3))
        vol = ndimage.grey_opening(vol, (7, 3, 3))

        # Threshold and connected components
        leotta_thresh = self.config.get('leotta_thresh', 0.3)
        vol_label, label_count = ndimage.label(vol > leotta_thresh)

        # Flatten over time, max label is a heuristic
        max_label = np.max(vol_label, 0)

        # count the number of pixels for each label in 2D

        # remove labels (back in 3D) that were too small
        if 0:
            idx, cnts = np.unique(max_label, return_counts=True)
            to_remove = idx[cnts < 50]
            vol_label[np.isin(vol_label, to_remove)] = 0

        # flatten again without the smaller blobs
        label_img = np.max(vol_label, 0)
        # idx, cnts = np.unique(max_label, return_counts=True)
        return label_img

    def _predict_crall(self):
        import rich
        from geowatch.utils import util_kwimage
        from scipy import ndimage
        import kwarray
        import numpy as np
        raw_heatmap = self.heatmap_thwc

        SHOW = 0
        scale_factor = self.config['scale_factor']
        thresh = self.config['thresh']
        if scale_factor == 'auto':
            scale_factor = 4

        _t, _h, _w, _c = self.heatmap_thwc.shape

        if self.bounds is not None:
            mask = self.bounds.to_mask(dims=(_h, _w)).data
        else:
            mask = None

        if self.config['viz_out_dir']:
            SHOW = 1

        if SHOW:
            import kwplot
            from geowatch.utils import util_kwplot
            dpath = self.config['viz_out_dir']
            finalizer = util_kwplot.FigureFinalizer(dpath, cropwhite=False)

        _n = [1]
        def PRINT_STEP(msg, _n=_n):
            import rich
            n = _n[0]
            rich.print(f'Step {n}: {msg}')
            _n[0] += 1

        rich.print(f'* Given: raw_heatmap.shape={raw_heatmap.shape}')

        # TODO: better downscaling?
        if scale_factor == 1:
            downscaled = raw_heatmap
        else:
            downscaled = raw_heatmap[:, ::scale_factor, ::scale_factor, :]
            PRINT_STEP(f'Downscale by {scale_factor}x to: {downscaled.shape}')
            downscaled = downscaled.copy()

        if mask is not None:
            small_mask = mask[::scale_factor, ::scale_factor]
        else:
            small_mask = None

        PRINT_STEP('Impute NaN')
        imputed = impute_nans2(downscaled)
        rich.print('... Finished Impute')

        if mask is not None:
            PRINT_STEP('Masking Small Heatmap')
            masked = imputed * small_mask[None, :, :, None]
        else:
            masked = imputed
        cube = FeatureCube(masked,
                           TimeIntervalSequence.coerce(self.heatmap_time_intervals),
                           self.classes)
        agg_cube = cube.window_max('12 months')

        if SHOW:
            fig, _ = kwplot.imshow(agg_cube.draw(), title='Input Feature Cube')
            finalizer.finalize(fig, f'step_{_n[0]:03d}_agg_cube.png')

        if 'ac_salient' in self.classes:
            cube1 = agg_cube.take_channels('ac_salient')
        else:
            cube1 = agg_cube.take_channels('salient')

        # cube2 = agg_cube.take_channels('No Activity')
        if 'Site Preparation' in self.classes:
            cube3 = agg_cube.take_channels('Site Preparation')
        else:
            cube3 = None

        if 'Active Construction' in self.classes:
            cube4 = agg_cube.take_channels('Active Construction')
        else:
            cube4 = None

        if 1:
            # hack for old model
            if cube4 is not None:
                mask = (cube1.heatmap_thwc == 0)
                cube1.heatmap_thwc[mask] = cube4.heatmap_thwc[mask]
        # cube5 = agg_cube.take_channels('Post Construction')

        if cube4 is None:
            cube8 = cube1
        else:
            cube8 = cube4 * cube1
        # cube6 = cube2 * cube1
        # cube7 = cube3 * cube1
        # cube9 = cube5 * cube1

        PRINT_STEP('Morphology')
        # cube1.heatmap_thwc = ndimage.grey_closing(cube1.heatmap_thwc, (3, 5, 5, 1))
        # cube1.heatmap_thwc = ndimage.grey_opening(cube1.heatmap_thwc, (3, 5, 5, 1))
        if cube4 is not None:
            cube4.heatmap_thwc = ndimage.grey_closing(cube4.heatmap_thwc, (3, 5, 5, 1))
            cube4.heatmap_thwc = ndimage.grey_opening(cube4.heatmap_thwc, (3, 5, 5, 1))

        # Spatial dilation
        cube1.heatmap_thwc = ndimage.grey_dilation(cube1.heatmap_thwc, (1, 5, 5, 1))

        if cube4 is not None:
            cube4.heatmap_thwc = ndimage.grey_dilation(cube4.heatmap_thwc, (1, 5, 5, 1))
        # cube8.heatmap_thwc = ndimage.grey_closing(cube8.heatmap_thwc, (3, 5, 5, 1))
        # cube8.heatmap_thwc = ndimage.grey_opening(cube8.heatmap_thwc, (3, 5, 5, 1))

        # cube2.heatmap_thwc = (cube2.heatmap_thwc > thresh).astype(np.float32)
        if cube3 is not None:
            cube3.heatmap_thwc = (cube3.heatmap_thwc > thresh).astype(np.float32)
        if cube4 is not None:
            cube4.heatmap_thwc = (cube4.heatmap_thwc > thresh).astype(np.float32)
        # cube5.heatmap_thwc = (cube5.heatmap_thwc > thresh).astype(np.float32)

        # cube6.heatmap_thwc = (cube6.heatmap_thwc > thresh).astype(np.float32)
        # cube7.heatmap_thwc = (cube7.heatmap_thwc > thresh).astype(np.float32)
        cube8.heatmap_thwc = (cube8.heatmap_thwc > thresh).astype(np.float32)
        # cube9.heatmap_thwc = (cube9.heatmap_thwc > thresh).astype(np.float32)

        if SHOW:
            fig2 = kwplot.figure(fnum=2, doclf=1)
            cubes = [cube1, cube4, cube8]
            ncubes = sum([c is not None for c in cubes])
            pnum = kwplot.PlotNums(nSubplots=ncubes)
            if cube1 is not None:
                kwplot.imshow(cube1.draw(), pnum=pnum(), fnum=2, title='Salient')
            if cube4 is not None:
                kwplot.imshow(cube4.draw(), pnum=pnum(), fnum=2, title='Morphed Active')
            if cube8 is not None:
                kwplot.imshow(cube8.draw(), pnum=pnum(), fnum=2, title='Binarized Active * Salient')
            finalizer.finalize(fig2, f'step_{_n[0]:03d}_cubes.png')

        PRINT_STEP('Max Saliency')
        max_saliency = cube1.heatmap_thwc.max(axis=0)[..., 0]
        max_saliency_stats = kwarray.stats_dict(max_saliency)

        if 0:
            print('max_saliency_stats = {}'.format(ub.urepr(max_saliency_stats, nl=1)))

        PRINT_STEP('Volume Labeling')
        vol_label, label_count = ndimage.label(cube8.heatmap_thwc > thresh)
        # Remove small seeds
        idx, cnts = np.unique(vol_label, return_counts=True)
        max_cnts = max(cnts)
        if max_cnts > 200:
            print(f'max_cnts = {ub.urepr(max_cnts, nl=1)}')
            count_thresh = min(50, max_cnts)
            to_remove = idx[cnts < count_thresh]
            n_remove = sum(to_remove)
            print(f'n_remove={n_remove}')
            vol_label[np.isin(vol_label, to_remove)] = 0

        labels1 = vol_label.max(axis=0)[..., 0]
        if SHOW:
            fig3, _ = kwplot.imshow(util_kwimage.colorize_label_image(labels1, label_to_color={0: 'black'}), fnum=3, doclf=1, title='Volume Labels')
            finalizer.finalize(fig3, f'step_{_n[0]:03d}_volume_labels.png')

        unique_labels = np.setdiff1d(np.unique(labels1), [0])

        if self.bounds is not None:
            self._intermediates['small_bounds'] = self.bounds.scale(1 / scale_factor)

            if SHOW:
                fig4, _ = kwplot.imshow(util_kwimage.colorize_label_image(labels1, label_to_color={0: 'black'}), fnum=4, doclf=1, title='Volume Labels With Bounds')
                self._intermediates['small_bounds'].draw(edgecolor='white', fill=0)
                finalizer.finalize(fig4, f'step_{_n[0]:03d}_volume_labels_bounds.png')

        do_watershed = self.bounds is not None
        if do_watershed:
            PRINT_STEP(f'Initialize {len(unique_labels)} Seed Points')
            markers = np.zeros_like(max_saliency, dtype=np.int32)
            marker_points = []
            for labelid in unique_labels:
                mask = labels1 == labelid
                idxmax = (max_saliency * mask).argmax()
                highval = np.unravel_index(idxmax, mask.shape)
                r, c = highval[0], highval[1]
                marker_points.append((c, r))
                markers[r, c] = labelid

            PRINT_STEP('Watershed')
            if SHOW:
                watershed_image = max_saliency
                fig6, _ = kwplot.imshow(watershed_image, fnum=6, doclf=1, title='Watershed Energy')
                import kwimage
                kwimage.Points(xy=np.array(marker_points)).draw()
                finalizer.finalize(fig6, f'step_{_n[0]:03d}_watershed_energy.png')

                ...
            from skimage.segmentation import watershed
            labels = watershed(-max_saliency, markers=markers, mask=small_mask)
        else:
            labels = labels1

        if SHOW:
            fig5, _ = kwplot.imshow(util_kwimage.colorize_label_image(labels, label_to_color={0: 'black'}), fnum=4, doclf=1, title='Final Labels')
            finalizer.finalize(fig5, f'step_{_n[0]:03d}_final_labels.png')
        return labels

    def predict(self):
        """
        Predict the spatial polygons

        Returns:
            ndarray:
                A single [H, W] integer map indicating a cluster label for each
                pixel. I.e. a spatial segmentation of the sites.
        """
        import einops
        import kwarray
        import kwimage
        import kwutil
        import numpy as np
        import rich
        import sklearn
        import sklearn.cluster
        import sklearn.decomposition

        _t, _h, _w, _c = self.heatmap_thwc.shape
        orig_dims = _t * _c

        if 1:
            # defaults
            max_dims = min(32, orig_dims)
            scale_factor = 2

        # Setup Config
        workers = kwutil.util_parallel.coerce_num_workers(self.config['workers'])

        algo = self.config['algo']

        if algo == 'leotta':
            return self._predict_leotta()

        if algo == 'crall':
            return self._predict_crall()

        if algo == 'dbscan':
            dbscan = sklearn.cluster.DBSCAN(
                eps=.009, min_samples=5, metric='cosine', metric_params=None,
                algorithm='auto', leaf_size=30, n_jobs=workers)
            cluster_algo = dbscan
            max_dims = min(8, orig_dims)
            scale_factor = 8
        elif algo == 'meanshift':
            mean_shift = sklearn.cluster.MeanShift(bandwidth=None, seeds=None,
                                                   bin_seeding=False,
                                                   # TODO: set based on resolution
                                                   min_bin_freq=1,
                                                   cluster_all=True,
                                                   n_jobs=workers,
                                                   max_iter=500)
            cluster_algo = mean_shift
            max_dims = min(32, orig_dims)
            scale_factor = 4
        elif algo == 'kmeans':
            from sklearn.cluster import MiniBatchKMeans
            mb_kmeans = MiniBatchKMeans(n_clusters=64, batch_size=4096,
                                        n_init='auto', random_state=1)
            cluster_algo = mb_kmeans
            scale_factor = 2
            max_dims = orig_dims
        else:
            raise KeyError(algo)

        pca = sklearn.decomposition.PCA(n_components=max_dims)
        reduce_dims_algo = pca

        rich.print("[green]--- Start PolygonExtractor Predict ---")
        rich.print
        rich.print('config = {}'.format(ub.urepr(self.config, nl=1)))
        rich.print('* reduce_dims_algo = {}'.format(ub.urepr(reduce_dims_algo, nl=1)))
        rich.print('* cluster_algo = {}'.format(ub.urepr(cluster_algo, nl=1)))

        # Run Algorithm

        def PRINT_STEP(msg, _n=[1]):
            import rich
            n = _n[0]
            rich.print(f'Step {n}: {msg}')
            _n[0] += 1

        raw_heatmap = self.heatmap_thwc

        if self.bounds is not None:
            mask = self.bounds.to_mask(dims=(_h, _w)).data
        else:
            mask = None

        rich.print(f'* Given: raw_heatmap.shape={raw_heatmap.shape}')

        # TODO: better downscaling?
        if scale_factor == 1:
            downscaled = raw_heatmap
        else:
            downscaled = raw_heatmap[:, ::scale_factor, ::scale_factor, :]
            PRINT_STEP(f'Downscale by {scale_factor}x to: {downscaled.shape}')
            downscaled = downscaled.copy()

        if mask is not None:
            small_mask = mask[::scale_factor, ::scale_factor]
        else:
            small_mask = None

        PRINT_STEP('Impute NaN')
        imputed = impute_nans2(downscaled)
        rich.print('... Finished Impute')

        if mask is not None:
            PRINT_STEP('Masking Small Heatmap')
            masked = imputed * small_mask[None, :, :, None]
        else:
            masked = imputed

        SMART_HACK = 0
        if SMART_HACK:
            # from skimage.segmentation import watershed
            # from skimage.feature import peak_local_max
            cx1 = salient_cidx = self.classes.index('ac_salient')  # NOQA
            cx2 = active_cidx = self.classes.index('Active Construction')  # NOQA
            cx3 = siteprep_cidx = self.classes.index('Site Preparation')  # NOQA
            salient_chan = c1 = masked[..., cx1]
            PRINT_STEP('Stacking Modulated Features')
            active_chan = c2 = masked[..., cx2]
            siteprep_chan = c3 = masked[..., cx3]
            c4 = c1 * c2
            c5 = c1 * c3
            to_stack = [c1, c2, c3, c4, c5]

            self._intermediates['salient_chan'] = salient_chan
            self._intermediates['siteprep_chan'] = siteprep_chan
            self._intermediates['active_chan'] = active_chan
            self._intermediates['small_bounds'] = self.bounds.scale(1 / scale_factor)

            small_heatmap = np.stack(to_stack, axis=3)
            rich.print(f'small_heatmap.shape={small_heatmap.shape}')

            PRINT_STEP('Modulate By Saliency')
            max_saliency = salient_chan.max(axis=0, keepdims=1)
            small_heatmap = small_heatmap

            max_saliency_2d = max_saliency[0, :, :]
        else:
            small_heatmap = masked
            max_saliency_2d = None

        if self.config.get('robust_normalize'):
            small_heatmap = kwarray.robust_normalize(small_heatmap)
            PRINT_STEP('Robust Normalize')
        else:
            small_heatmap = small_heatmap
            PRINT_STEP('Skip Robust Normalize')

        t, h, w, c = small_heatmap.shape

        X = einops.rearrange(small_heatmap, 't h w c -> (h w) (t c)')
        PRINT_STEP(f'Rearange (combine time / channels) to: X.shape={X.shape}')

        Xhat = reduce_dims_algo.fit_transform(X)
        PRINT_STEP(f'Reduce dimensionality: Xhat.shape={Xhat.shape}')

        if self.config.get('positional_encoding'):
            scale = self.config.get('positional_encoding_scale', 1)
            rr, cc = np.meshgrid(
                np.linspace(-scale, scale, h),
                np.linspace(-scale, scale, w)
            )
            Xhat2 = np.concatenate([Xhat, rr.T.ravel()[:, None], cc.T.ravel()[:, None]], axis=1)
            PRINT_STEP(f'* Append Positional Encoding. Xhat2.shape={Xhat2.shape}')
        else:
            Xhat2 = Xhat

        PRINT_STEP('Run Clustering Algorithm')
        yhat = cluster_algo.fit_predict(Xhat2) + 1
        rich.print('... Finished Clustering Algorithm')

        small_label_img = einops.rearrange(yhat, '(h w) -> h w', w=w, h=h)
        PRINT_STEP(f'* Convert back to spatial arangement: small_label_img.shape={small_label_img.shape}')
        small_label_img = np.ascontiguousarray(small_label_img).astype(np.uint8)

        if mask is not None:
            PRINT_STEP('Masking label img')
            small_label_img = small_mask * small_label_img

        if max_saliency_2d is not None:
            saliency_mask = max_saliency_2d > 0.2
            small_label_img = saliency_mask * small_label_img
            # coords = peak_local_max(max_saliency_2d, footprint=np.ones((7, 7)), labels=saliency_mask)
            # lbl = watershed(max_saliency_2d, mask=saliency_mask)
            if 0:
                from geowatch.utils import util_kwimage
                import kwplot
                # canvas = util_kwimage.colorize_label_image(lbl)
                kwplot.imshow(max_saliency_2d, fnum=1, doclf=1)
                kwplot.imshow(saliency_mask, fnum=1, doclf=1)
                kwplot.imshow(util_kwimage.colorize_label_image(small_label_img, label_to_color={0: 'black'}))

        # PRINT_STEP('Morphology on Label')
        # small_label_img = kwimage.morphology(
        #     small_label_img, 'close', kernel=5, element='ellipse')

        if 0:
            small_feat = Xhat2.reshape(h, w, -1)
            ohe = kwarray.one_hot_embedding(small_label_img, small_label_img.max() + 1, dim=0)

            superpixel_feats = []
            superpixel_polys = []
            for _ohe in ohe[1:]:
                poly = kwimage.Mask(_ohe.astype(np.uint8), 'c_mask')
                poly = poly.to_multi_polygon(pixels_are='areas')

                poly_feat = (_ohe[:, :, None] * small_feat).sum(axis=(0, 1)) / (_ohe.sum() + 1)
                superpixel_polys.append(poly)
                superpixel_feats.append(poly_feat)

            keep_superpixel_idxs = np.arange(1, len(ohe))
            superpixel_feats = np.array(superpixel_feats)
            flags = superpixel_feats.max(axis=1) >= 0.000
            keep_superpixel_idxs = keep_superpixel_idxs[flags]
            superpixel_feats = superpixel_feats[flags]
            superpixel_polys = list(ub.compress(superpixel_polys, flags))

            dist = ((superpixel_feats[:, None, :] - superpixel_feats[None, :, :]) ** 2).sum(axis=2)
            ii, jj = np.where((dist < 3) & (dist > 0))
            import networkx as nx
            g = nx.Graph()
            g.add_edges_from(zip(ii, jj))

            new_small_label_img = np.zeros_like(small_label_img)

            next_new_id = 1
            for _, cc in enumerate(nx.connected_components(g), start=1):
                cc_polys = list(ub.take(superpixel_polys, cc))
                from shapely.ops import unary_union
                try:
                    cc_poly = unary_union([p.to_shapely().buffer(1) for p in cc_polys])
                except Exception:
                    continue
                for poly in kwimage.MultiPolygon.from_shapely(cc_poly):
                    poly.fill(new_small_label_img, value=next_new_id, pixels_are='areas')
                    next_new_id += 1

            if 0:
                kwplot.imshow(util_kwimage.colorize_label_image(new_small_label_img, label_to_color={0: 'black'}))

            small_label_img = new_small_label_img

        label_img = kwimage.imresize(small_label_img, scale=(scale_factor, scale_factor), interpolation='nearest')
        PRINT_STEP(f'* Step 8. Resize back to full scale: label_img.shape={label_img.shape}')

        rich.print("[green]--- End PolygonExtractor Predict ---")

        self._intermediates = {
            'max_saliency_2d': max_saliency_2d,
            'small_label_img': small_label_img,
            'Xhat2': Xhat2,
        }

        if 0:
            import kwplot
            kwplot.autompl()
            from geowatch.utils import util_kwimage  # NOQA
            canvas = util_kwimage.colorize_label_image(label_img, label_to_color={0: 'black'})
            kwplot.imshow(canvas, fnum=1, doclf=1)

            feature_pca = util_kwimage.ensure_false_color(
                Xhat2.reshape(h, w, -1), method='ortho')
            kwplot.imshow(feature_pca, fnum=2, doclf=1)

        return label_img

    def predict_polygons(self, return_info=False):
        label_img = self.predict()
        label_mask = LabelMask(label_img)
        polys = label_mask.to_multi_polygons()

        if self.config['polygon_simplify_tolerance'] is not None:
            # TODO: handle difference between pixel and world units.
            simplify_thresh = int(self.config['polygon_simplify_tolerance'])
            polys = [p.simplify(simplify_thresh) for p in polys]

        if return_info:
            info = {}
            info['label_mask'] = label_mask
            return polys, info
        else:
            return polys

    def show(self):
        import kwplot
        kwplot.autompl()
        stacked = self.draw_timesequence()
        kwplot.imshow(stacked)

    def draw_intermediate(self):
        import kwimage
        import numpy as np
        from geowatch.utils import util_kwimage
        salient = self._intermediates['salient_chan']
        # active = self._intermediates['active_chan']
        # siteprep = self._intermediates['siteprep_chan']
        small_bounds = self._intermediates['small_bounds']

        # heatmaps = np.stack([salient, active, siteprep], axis=3)
        heatmaps = np.stack([salient], axis=3)
        heatmaps = (heatmaps > 0.5).astype(np.float32)

        name_to_color = {cat['name']: kwimage.Color.coerce(cat['color']).as01()
                          for cat in self.classes.cats.values()}
        channel_colors = [name_to_color[n] for n in [
            'ac_salient', 'Site Preparation', 'Active Construction']]
        channel_colors = ['red']

        to_show_frames = [kwimage.nodata_checkerboard(
            util_kwimage.perchannel_colorize(h.astype(np.float32), channel_colors=channel_colors)
        ) for h in heatmaps]

        if self.bounds is not None:
            to_show_frames = [small_bounds.draw_on(frame, edgecolor='white', fill=False) for frame in to_show_frames]

        stacked = kwimage.stack_images_grid(to_show_frames, pad=10, bg_value='kitware_green')
        return stacked

    def draw_timesequence(self):
        import kwimage
        import numpy as np
        import kwarray
        from geowatch.utils import util_kwimage
        heatmaps = self.heatmap_thwc
        norm_heatmaps = kwarray.robust_normalize(heatmaps)

        if self.classes is not None:
            channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                              for cat in self.classes.cats.values()]
            to_show_frames = [kwimage.nodata_checkerboard(
                util_kwimage.perchannel_colorize(h.astype(np.float32), channel_colors=channel_colors)
            ) for h in norm_heatmaps]
        else:
            to_show_frames = [kwimage.nodata_checkerboard(util_kwimage.ensure_false_color(h.astype(np.float32)))
                              for h in norm_heatmaps]
        if self.bounds is not None:
            to_show_frames = [self.bounds.draw_on(frame, edgecolor='white', fill=False) for frame in to_show_frames]
        stacked = kwimage.stack_images_grid(to_show_frames, pad=10, bg_value='kitware_green')
        return stacked

    @classmethod
    def demo(cls, **kwargs):
        """
        Create an instance of the problem on toy data.

        Args:
            **kwargs: passed to :func:`PolygonExtractor.demodata`

        Returns:
            PolygonExtractor
        """
        kwargs = cls.demodata(**kwargs)
        self = cls(**kwargs)
        return self

    @classmethod
    def demodata(cls, real_categories=False, rng=0):
        """
        Create toydata to test and demo the API

        Args:
            real_categories (bool): if False, use fake cateogires

            rng (int | None):
                random seed, or None to use global seed.

        Returns:
            Dict[str, Any]: A dictionary that can be used as kwargs
               to construct and instance of this class.
        """
        import kwimage
        import numpy as np
        import kwcoco
        import kwarray

        rng = kwarray.ensure_rng(rng)

        H, W = 128, 128
        dsize = np.array([W, H])

        # Define a random polygon and create several copies of it.
        poly0 = kwimage.Polygon.random(rng=rng)
        poly0 = poly0.translate((-0.5, -0.5))

        scale_factor = dsize / 4

        poly0 = poly0.scale(scale_factor)
        poly0 = poly0.translate((W / 2, H / 2))

        shift = poly0.box().width * 0.65

        poly1 = poly0.translate((-shift, -shift))
        poly2 = poly0.translate((+shift, +shift))
        poly3 = poly0.translate((-shift, +shift))
        poly4 = poly0.translate((+shift, -shift))

        polys = {
            'poly0': poly0,
            'poly1': poly1,
            'poly2': poly2,
            'poly3': poly3,
            'poly4': poly4,
        }

        # Use bounds of all polys to define a "bounds" object
        bounds = None
        for poly in polys.values():
            if bounds is None:
                bounds = poly
            else:
                bounds = bounds.union(poly)
        bounds = bounds.convex_hull
        bounds = bounds.scale(1.1, about='centroid')

        if real_categories:
            classes = kwcoco.CategoryTree.from_coco([
                {'id': 10, 'name': 'No Activity', 'color': 'tomato'},
                {'id': 11, 'name': 'Site Preparation', 'color': 'gold'},
                {'id': 12, 'name': 'Active Construction', 'color': 'lime'},
                {'id': 13, 'name': 'Post Construction', 'color': 'darkturquoise'},
                {'id': 14, 'name': 'ac_salient', 'color': 'white'},
            ])
        else:
            classes = kwcoco.CategoryTree.from_coco([
                {'id': 20, 'name': 'Red', 'color': 'red'},
                {'id': 21, 'name': 'Green', 'color': 'green'},
                {'id': 22, 'name': 'Blue', 'color': 'blue'},
            ])

        # Define a simple timeseries for each polygon to follow
        # TODO: there might be a nicer way to encode this to allow demodata to
        # capture more aspects of the problem.
        sequences = {
            'poly0': [
                {'frame': 2, 'class': 0},
                {'frame': 3, 'class': 0},
                {'frame': 4, 'class': 0},
                {'frame': 5, 'class': 0},
                {'frame': 6, 'class': 0},
                {'frame': 7, 'class': 0},
                {'frame': 8, 'class': 1},
                {'frame': 9, 'class': 2},
            ],

            'poly1': [
                {'frame': 4, 'class': 0},
                {'frame': 5, 'class': 1},
                {'frame': 6, 'class': 1},
                {'frame': 7, 'class': 1},
                {'frame': 8, 'class': 2},
                {'frame': 9, 'class': 2},
                {'frame': 10, 'class': 2},
            ],

            'poly2':  [
                {'frame': 1, 'class': 0},
                {'frame': 2, 'class': 1},
                {'frame': 3, 'class': 1},
                {'frame': 4, 'class': 1},
                {'frame': 5, 'class': 2},
            ],

            'poly3': [
                {'frame': 12, 'class': 1},
                {'frame': 13, 'class': 2},
                {'frame': 14, 'class': 2},
            ],

            'poly4': [
                {'frame': 4, 'class': 0},
                {'frame': 5, 'class': 1},
                {'frame': 6, 'class': 1},
                {'frame': 7, 'class': 1},
                {'frame': 8, 'class': 2},
                {'frame': 9, 'class': 2},
                {'frame': 10, 'class': 2},
            ],
        }

        # Create a heatmap time sequence and render polygons onto it
        C = len(classes)
        frame_shape = (H, W, C)
        # canvas = np.zeros(frame_shape, dtype=np.float32)
        canvas = rng.rand(*frame_shape).astype(np.float32) * 0.01

        heatmap_frames = [canvas.copy() for _ in range(16)]

        import kwutil
        base = kwutil.util_time.datetime.coerce('2020-01-01')
        delta = kwutil.util_time.timedelta.coerce('2 weeks')
        heatmap_time_intervals = [
            (base + (delta * idx), base + (delta * (idx + 1)))
            for idx in range(len(heatmap_frames))
        ]
        heatmap_time_intervals = [
            TimeInterval.coerce(d) for d in heatmap_time_intervals
        ]

        for key in polys.keys():
            poly = polys[key]
            seq = sequences[key]
            for info in seq:
                idx = info['frame']
                cidx = info['class']
                chan = np.ascontiguousarray(heatmap_frames[idx][..., cidx])
                new_chan = poly.fill(chan, value=1)
                heatmap_frames[idx][..., cidx] = new_chan
                if real_categories:
                    # hack to add in saliency for active classes
                    if classes[cidx] in {'Active Construction', 'Site Preparation'}:
                        cidx = classes.index('ac_salient')
                        chan = np.ascontiguousarray(heatmap_frames[idx][..., cidx])
                        new_chan = poly.fill(chan, value=1)
                        heatmap_frames[idx][..., cidx] = new_chan

        heatmap_thwc = np.array(heatmap_frames)
        kwargs = {
            'bounds': bounds,
            'heatmap_thwc': heatmap_thwc,
            'classes': classes,
            'heatmap_time_intervals': heatmap_time_intervals,
        }
        return kwargs


def impute_nans(data):
    '''
    interpolate to fill nan values

    TODO: tests

    data = np.random.rand(5, 3, 3, 2)
    data[data < 0.1] = np.nan

    data = np.random.rand(54, 295, 296, 5)
    data[data < 0.1] = np.nan

    import timerit
    ti = timerit.Timerit(1, bestof=1, verbose=3)

    for timer in ti.reset('impute_nans2'):
        with timer:
            impute_nans2(data)

    for timer in ti.reset('impute_nans'):
        with timer:
            impute_nans(data)
    '''
    from scipy.interpolate import NearestNDInterpolator
    import numpy as np
    mask = np.isfinite(data)
    valid_points = np.stack(np.where(mask), axis=1)
    invalid_points = np.stack(np.where(~mask), axis=1)
    valid_data = data[mask].ravel()

    if len(data.shape) == 4:
        t, h, w, c = data.shape
        # Hack: scale space to force interpolation over time first space second,
        # and channel last. Ideally we would never interpolate over space or
        # channel, but not sure how to do that atm. Is there a way to only fill in
        # values over the time axis?
        multiplier = np.array([1 / t, 10 * t, 10 * t, t * w * h * 10])[None, :]
    elif len(data.shape) == 3:
        t, h, w = data.shape
        multiplier = np.array([1 / t, 10 * t, 10 * t])[None, :]

    f_nearest = NearestNDInterpolator(valid_points * multiplier, valid_data)
    fill_data = f_nearest(invalid_points * multiplier)

    new_data = data.copy()
    new_data[tuple(invalid_points.T)] = fill_data
    return new_data


def impute_nans2(data):
    import numpy as np
    bad_value_mask = ~np.isfinite(data)
    new_data = data.copy()
    frame_new_prev = None
    for idx in range(data.shape[0]):
        frame_mask = bad_value_mask[idx]
        frame_new = new_data[idx]
        if frame_new_prev is None:
            frame_new[frame_mask] = 0
        else:
            # Impute values from previous frame
            frame_new[frame_mask] = frame_new_prev[frame_mask]
        frame_new_prev = frame_new
    return new_data


class FeatureCube(ub.NiceRepr):
    """
    Container for a [T, H, W, C] heatmap and corresponding T-lengthed time
    intervals.

    Ignore:
        self = FeatureCube.demo()
        print(f'self={self}')
        new = self.window_max('1 month')
        self.heatmap_time_intervals
        new.heatmap_time_intervals

        channels = ['ac_salient']

        import kwplot
        kwplot.autompl()
        kwplot.imshow(self.draw(), pnum=(1, 3, 1))
        kwplot.imshow(new.draw(), pnum=(1, 3, 2))

        new2 = new.take_channels(['ac_salient'])
        kwplot.imshow(new2.draw(), pnum=(1, 3, 3))
    """
    def __init__(self, heatmap_thwc, heatmap_time_intervals=None, classes=None):
        self.heatmap_thwc = heatmap_thwc
        self.heatmap_time_intervals = heatmap_time_intervals
        self.classes = classes

    def __nice__(self):
        return f'{self.heatmap_thwc.shape}'

    def take_channels(self, channels):
        if isinstance(channels, str):
            import kwcoco
            channels = kwcoco.FusedChannelSpec.coerce(channels).as_list()
        chan_idxs = [self.classes.index(c) for c in channels]
        new_heatmaps = self.heatmap_thwc[..., chan_idxs]

        new_graph = self.classes.graph.subgraph(channels)
        new_classes = self.classes.__class__(new_graph)
        new = FeatureCube(new_heatmaps, self.heatmap_time_intervals, classes=new_classes)
        return new

    def window_max(self, window):
        import numpy as np
        groupxs = self.heatmap_time_intervals._compute_time_window(window)
        new_intervals = self.heatmap_time_intervals.apply_grouping(groupxs)
        new_heatmaps = []
        for idxs in groupxs:
            combo = np.nanmax(self.heatmap_thwc[idxs], axis=0)
            new_heatmaps.append(combo)
        new_heatmaps = np.array(new_heatmaps)
        new = FeatureCube(new_heatmaps, new_intervals, classes=self.classes)
        return new

    @classmethod
    def demo(self):
        p = PolygonExtractor.demo(real_categories=1)
        intervals = TimeIntervalSequence.coerce(p.heatmap_time_intervals)
        self = FeatureCube(p.heatmap_thwc, intervals, p.classes)
        return self

    def draw(self):
        import kwimage
        import numpy as np
        import kwarray
        from geowatch.utils import util_kwimage
        heatmaps = self.heatmap_thwc
        norm_heatmaps = kwarray.robust_normalize(heatmaps)

        if self.classes is not None:
            from geowatch import heuristics
            heuristics.ensure_heuristic_category_tree_colors(self.classes)

            channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                              for cat in self.classes.cats.values()]
            to_show_frames = [kwimage.nodata_checkerboard(
                util_kwimage.perchannel_colorize(h.astype(np.float32), channel_colors=channel_colors)
            ) for h in norm_heatmaps]
        else:
            to_show_frames = [kwimage.nodata_checkerboard(util_kwimage.ensure_false_color(h.astype(np.float32)))
                              for h in norm_heatmaps]
        # if self.bounds is not None:
        #     to_show_frames = [self.bounds.draw_on(frame, edgecolor='white', fill=False) for frame in to_show_frames]
        stacked = kwimage.stack_images_grid(to_show_frames, pad=10, bg_value='kitware_green')
        return stacked

    def __mul__(self, other):
        new_heatmap = self.heatmap_thwc * other.heatmap_thwc
        new = FeatureCube(new_heatmap, self.heatmap_time_intervals, self.classes)
        return new


class TimeIntervalSequence(list):
    """
    A list of non-overlapping time intervals

    Ignore:
        from geowatch.tasks.tracking.polygon_extraction import *  # NOQA
        a = TimeInterval.coerce(('2020-01-01', '2020-02-01'))
        b = TimeInterval.coerce(('2020-02-01', '2020-03-01'))
        c = TimeInterval.coerce(('2020-03-01', '2020-03-15'))
        d = TimeInterval.coerce(('2020-03-07', '2020-04-01'))
        e = TimeInterval.coerce(('2020-04-15', '2020-05-01'))
        seq = TimeIntervalSequence([a, b, c, d, e])
        seq._compute_time_window('1 month')
    """

    def apply_grouping(self, groupxs):
        from functools import reduce
        new = TimeIntervalSequence()
        for idxs in groupxs:
            part = list(ub.take(self, idxs))
            new_part = reduce(TimeInterval.union, part[1:], part[0]).enclosure
            new.append(new_part)
        return new

    @classmethod
    def coerce(self, data):
        if isinstance(data, TimeIntervalSequence):
            self = data
        else:
            self = TimeIntervalSequence([TimeInterval.coerce(d) for d in data])
        return self

    def _compute_time_window(self, window):
        """
        Example:
            >>> window = '7days'
            >>> from kwutil import util_time
            >>> self = heatmap_dates = TimeIntervalSequence(map(TimeInterval.coerce, [
            >>>     '2020-01-01', '2020-01-02', '2020-02-01',
            >>>     '2020-02-02', '2020-03-14', '2020-03-23',
            >>>     '2020-04-01', '2020-06-23', '2020-06-26',
            >>>     '2020-06-27', '2020-06-28', ]))
            >>> groupxs = heatmap_dates._compute_time_window(window)
            >>> new = self.apply_grouping(groupxs)
            >>> print(f'groupxs={groupxs}')
            >>> print(f'new={new}')
            >>> groupxs = heatmap_dates._compute_time_window(None)
            >>> new = self.apply_grouping(groupxs)
            >>> print(f'groupxs={groupxs}')
            >>> print(f'new={new}')
            >>> groupxs = heatmap_dates._compute_time_window(3)
            >>> new = self.apply_grouping(groupxs)
            >>> print(f'groupxs={groupxs}')
            >>> print(f'new={new}')
        """
        import kwarray
        from kwutil import util_time
        import numpy as np
        num_frames = len(self)
        if window is None:
            bucket_ids = np.arange(num_frames)
        elif isinstance(window, str):
            delta = util_time.coerce_timedelta(window).total_seconds()
            start_unixtimes = np.array([x.start.timestamp() for x in self])
            start_unixtimes = start_unixtimes - start_unixtimes[0]
            bucket_ids = (start_unixtimes // delta).astype(int)
        elif isinstance(window, int):
            assert num_frames is not None
            frame_indexes = np.arange(num_frames)
            bucket_ids = frame_indexes // window
        else:
            raise NotImplementedError('')
        unique_ids, groupxs = kwarray.group_indices(bucket_ids)
        return groupxs


class TimeInterval(portion.Interval):
    """
    Represents an interval in time.

    Ignore:
        import liberator
        lib = liberator.Liberator()
        lib.add_dynamic(portion.Interval)
        lib.expand(['portion'])
        print(lib.current_sourcecode())

    Example:
        from geowatch.tasks.tracking.polygon_extraction import *  # NOQA
        a = TimeInterval.coerce(('2020-01-01', '2020-02-01'))
        b = TimeInterval.coerce(('2020-02-01', '2020-03-01'))
        c = TimeInterval.coerce(('2020-03-01', '2020-03-15'))
        d = TimeInterval.coerce(('2020-03-07', '2020-04-01'))
        e = TimeInterval.coerce(('2020-04-15', '2020-05-01'))
    """

    @property
    def start(self):
        return self.lower

    @property
    def stop(self):
        return self.upper

    @classmethod
    def coerce(TimeInterval, data):
        import kwutil
        if isinstance(data, TimeInterval):
            self = data
        elif isinstance(data, (str, kwutil.util_time.datetime_cls, float, int)):
            start = kwutil.util_time.datetime.coerce(data)
            stop = start
            self = TimeInterval.closed(start, stop)
        elif ub.iterable(data) and len(data) == 2:
            start, stop = data
            start = kwutil.util_time.datetime.coerce(start)
            stop = kwutil.util_time.datetime.coerce(stop)
            self = TimeInterval.closed(start, stop)
        else:
            raise TypeError
        return self

    @classmethod
    def closed(TimeInterval, start, stop=None):
        """
        """
        if start is None:
            start = -float('inf')
        if stop is None:
            start = -float('inf')
        self = TimeInterval.from_atomic(portion.CLOSED, start, stop, portion.CLOSED)
        return self

    @classmethod
    def random(TimeInterval):
        import kwutil
        a = kwutil.util_time.datetime.random()
        b = kwutil.util_time.datetime.random()
        a, b = sorted([a, b])
        self = TimeInterval.closed(a, b)
        return self


def toydata_demo():
    """
    Example:
        from geowatch.tasks.tracking.polygon_extraction import *  # NOQA
        cls = PolygonExtractor(heatmaps, bounds)
        self = PolygonExtractor.demo()
    """
    from geowatch.utils import util_kwimage
    from geowatch.utils import util_kwplot
    import kwplot

    util_kwplot.autompl2()

    self = PolygonExtractor.demo()
    stacked = self.draw_timesequence()
    label_img = self.predict()

    canvas = util_kwimage.colorize_label_image(label_img)
    kwplot.imshow(stacked, pnum=(1, 2, 1))
    kwplot.imshow(canvas, pnum=(1, 2, 2))

    if 0:
        mean_img = self.heatmap_thwc.mean(axis=0)
        import kwplot
        kwplot.plt
        kwplot.imshow(mean_img, fnum=2)


def real_data_demo_case_1():
    from geowatch.utils import util_girder
    from geowatch.utils import util_kwimage
    import pickle
    import kwimage
    import kwarray
    import kwplot
    import numpy as np

    api_url = 'https://data.kitware.com/api/v1'
    heatmap_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d1423fdc1508d8bec1bc3', hash_prefix='f9eb14a7c0ec99'))

    truth_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d2624fdc1508d8bec1bca', hash_prefix='c164037856d2ce'))

    bas_bounds_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d29f6fdc1508d8bec1bcd', hash_prefix='9c555b1adc518b'))

    heatmap_dates_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d2b11fdc1508d8bec1bd2', hash_prefix='4cdcdf2a40a819'))

    true_info_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d4f35fdc1508d8bec1c86', hash_prefix='4a68c596804200'))

    truth_labels = pickle.loads(truth_fpath.read_bytes())
    heatmap_thwc = pickle.loads(heatmap_fpath.read_bytes())
    bounds = pickle.loads(bas_bounds_fpath.read_bytes())
    heatmap_time_intervals = pickle.loads(heatmap_dates_fpath.read_bytes())
    truth_info = pickle.loads(true_info_fpath.read_bytes())

    start_time = heatmap_time_intervals[0][0]
    end_time = heatmap_time_intervals[-1][1]
    print('start_time = {}'.format(ub.urepr(start_time, nl=1)))
    print('end_time = {}'.format(ub.urepr(end_time, nl=1)))

    import kwcoco
    classes = kwcoco.CategoryTree.from_coco([
        {'name': 'No Activity', 'color': 'tomato'},
        {'name': 'Site Preparation', 'color': 'gold'},
        {'name': 'Active Construction', 'color': 'lime'},
        {'name': 'Post Construction', 'color': 'darkturquoise'},
        {'name': 'ac_salient', 'color': 'white'},
    ])
    channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                      for cat in classes.cats.values()]

    unique_labels = np.unique(truth_labels)
    unique_colors = kwimage.Color.distinct(len(unique_labels))
    label_to_color = ub.dzip(unique_labels, unique_colors)
    label_to_color[0] = kwimage.Color.coerce('black').as01()

    truth_canvas = util_kwimage.colorize_label_image(truth_labels, label_to_color=label_to_color)

    frame0_canvas = kwarray.robust_normalize(heatmap_thwc[0])
    frame0_canvas = util_kwimage.perchannel_colorize(frame0_canvas, channel_colors=channel_colors)
    frame0_canvas = util_kwimage.ensure_false_color(frame0_canvas)
    frame0_canvas = kwimage.nodata_checkerboard(frame0_canvas)
    frame0_canvas = bounds.draw_on(frame0_canvas, fill=False, edgecolor='white')

    frameN_canvas = kwarray.robust_normalize(heatmap_thwc[-1])
    frameN_canvas = util_kwimage.perchannel_colorize(frameN_canvas, channel_colors=channel_colors)
    frameN_canvas = kwimage.nodata_checkerboard(frameN_canvas)
    frameN_canvas = bounds.draw_on(frameN_canvas, fill=False, edgecolor='white')

    truth_canvas = bounds.draw_on(truth_canvas, fill=False, edgecolor='white')

    # Color truth by status
    import numpy as np
    from geowatch import heuristics
    truth_canvas2 = kwimage.atleast_3channels(np.zeros_like(truth_labels)).astype(np.uint8) + 255
    status_to_color = {}
    for item in ub.ProgIter(list(truth_info.values()), desc='draw truth'):
        poly = kwimage.MultiPolygon.coerce(item['geometry_pxl'])
        color = heuristics.IARPA_STATUS_TO_INFO[item['status']]['color']
        edgecolor = kwimage.Color.random()
        truth_canvas2 = poly.draw_on(truth_canvas2, facecolor=color, edgecolor=edgecolor)
        status_to_color[item['status']] = color
    legend_img = kwplot.make_legend_img(status_to_color)
    truth_canvas2 = kwimage.stack_images([truth_canvas2, legend_img], axis=1)

    kwplot.autompl()
    kwplot.imshow(frame0_canvas, title=f'Frame 0 Colorized: {start_time}', fnum=1)
    kwplot.imshow(frameN_canvas, title=f'Frame N Colorized: {end_time}', fnum=2)
    kwplot.imshow(truth_canvas, title='Truth Labels', fnum=3)
    kwplot.imshow(truth_canvas2, title='Truth Status', fnum=4)

    self = PolygonExtractor(heatmap_thwc, bounds=bounds,
                            heatmap_time_intervals=heatmap_time_intervals,
                            classes=classes)
    stacked = self.draw_timesequence()
    kwplot.imshow(stacked, pnum=(1, 2, 1), fnum=5)

    label_img = self.predict()
    # label_img = self._predict_leotta()

    canvas = util_kwimage.colorize_label_image(label_img)
    kwplot.imshow(canvas, pnum=(1, 2, 2), fnum=5)


def real_data_demo_case_2():
    from geowatch.utils import util_girder
    from geowatch.utils import util_kwimage
    import pickle
    import kwimage
    import kwarray
    import kwplot
    import numpy as np

    api_url = 'https://data.kitware.com/api/v1'
    data_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654d6ce3fdc1508d8bec1c89', hash_prefix='d1425c26580e03'))

    data = pickle.loads(data_fpath.read_bytes())
    heatmap_thwc = data['heatmaps_thwc']
    bounds = kwimage.MultiPolygon.from_shapely(data['bas_gdf'].geometry.unary_union)
    heatmap_time_intervals = data['heatmap_time_intervals']
    truth_gdf = data['truth_gdf']

    t, h, w, c = heatmap_thwc.shape

    truth_colorized = np.ones((h, w, 3))
    truth_labels = np.zeros((h, w), dtype=np.int32)
    idx = 1
    for _, row in truth_gdf.iterrows():
        true_poly = kwimage.MultiPolygon.coerce(row['geometry'])
        truth_colorized = true_poly.draw_on(truth_colorized, color=row['color'])
        if row['status'] not in {'negative', 'ignore'}:
            truth_labels = true_poly.fill(truth_labels, value=idx)
            idx += 1

    truth_labels
    start_time = heatmap_time_intervals[0][0]
    end_time = heatmap_time_intervals[-1][1]
    print('start_time = {}'.format(ub.urepr(start_time, nl=1)))
    print('end_time = {}'.format(ub.urepr(end_time, nl=1)))

    import kwcoco
    classes = kwcoco.CategoryTree.from_coco([
        {'id': 30, 'name': 'No Activity', 'color': 'tomato'},
        {'id': 31, 'name': 'Site Preparation', 'color': 'gold'},
        {'id': 32, 'name': 'Active Construction', 'color': 'lime'},
        {'id': 33, 'name': 'Post Construction', 'color': 'darkturquoise'},
        {'id': 34, 'name': 'ac_salient', 'color': 'white'},
    ])
    print(list(classes))
    list(classes.graph.nodes())
    channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                      for cat in classes.cats.values()]

    unique_labels = np.unique(truth_labels)
    unique_colors = kwimage.Color.distinct(len(unique_labels))
    label_to_color = ub.dzip(unique_labels, unique_colors)
    label_to_color[0] = kwimage.Color.coerce('black').as01()

    truth_canvas = util_kwimage.colorize_label_image(truth_labels, label_to_color=label_to_color)

    frame0_canvas = kwarray.robust_normalize(heatmap_thwc[0])
    frame0_canvas = util_kwimage.perchannel_colorize(frame0_canvas, channel_colors=channel_colors)
    frame0_canvas = util_kwimage.ensure_false_color(frame0_canvas)
    frame0_canvas = kwimage.nodata_checkerboard(frame0_canvas)
    frame0_canvas = bounds.draw_on(frame0_canvas, fill=False, edgecolor='white')

    frameN_canvas = kwarray.robust_normalize(heatmap_thwc[-1])
    frameN_canvas = util_kwimage.perchannel_colorize(frameN_canvas, channel_colors=channel_colors)
    frameN_canvas = kwimage.nodata_checkerboard(frameN_canvas)
    frameN_canvas = bounds.draw_on(frameN_canvas, fill=False, edgecolor='white')

    truth_canvas = bounds.draw_on(truth_canvas, fill=False, edgecolor='white')

    # Color truth by status
    # import numpy as np
    # from geowatch import heuristics
    # truth_canvas2 = kwimage.atleast_3channels(np.zeros_like(truth_labels)).astype(np.uint8) + 255
    # status_to_color = {}
    # for item in ub.ProgIter(list(truth_info.values()), desc='draw truth'):
    #     poly = kwimage.MultiPolygon.coerce(item['geometry_pxl'])
    #     color = heuristics.IARPA_STATUS_TO_INFO[item['status']]['color']
    #     edgecolor = kwimage.Color.random()
    #     truth_canvas2 = poly.draw_on(truth_canvas2, facecolor=color, edgecolor=edgecolor)
    #     status_to_color[item['status']] = color
    # legend_img = kwplot.make_legend_img(status_to_color)
    # truth_canvas2 = kwimage.stack_images([truth_canvas2, legend_img], axis=1)

    kwplot.autompl()
    kwplot.imshow(frame0_canvas, title=f'Frame 0 Colorized: {start_time}', fnum=1)
    kwplot.imshow(frameN_canvas, title=f'Frame N Colorized: {end_time}', fnum=2)
    kwplot.imshow(truth_canvas, title='Truth Labels', fnum=3)
    kwplot.imshow(truth_colorized, title='Truth Status', fnum=4)

    self = PolygonExtractor(heatmap_thwc, bounds=bounds,
                            heatmap_time_intervals=heatmap_time_intervals,
                            classes=classes, config={
                                # 'algo': 'meanshift',
                            })

    # label_img = self.predict()
    # label_img = self._predict_leotta()
    label_img = self._predict_crall()

    canvas = util_kwimage.colorize_label_image(label_img)
    kwplot.imshow(canvas, pnum=(1, 2, 2), fnum=6)

    stacked = self.draw_timesequence()
    kwplot.imshow(stacked, pnum=(1, 2, 1), fnum=6)


def real_data_demo_case_3():
    from geowatch.utils import util_girder
    from geowatch.utils import util_kwimage
    import pickle
    import kwimage
    import kwarray
    import kwplot
    import numpy as np

    api_url = 'https://data.kitware.com/api/v1'
    data_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654e77f4314693d6b1df2655', hash_prefix='1b91956803365b51'))

    data = pickle.loads(data_fpath.read_bytes())
    heatmap_thwc = data['heatmap_thwc']
    bounds = kwimage.MultiPolygon.from_shapely(data['bas_gdf'].geometry.unary_union)
    heatmap_time_intervals = data['heatmap_time_intervals']
    truth_gdf = data['truth_gdf']

    t, h, w, c = heatmap_thwc.shape

    truth_colorized = np.ones((h, w, 3))
    truth_labels = np.zeros((h, w), dtype=np.int32)
    idx = 1
    for _, row in truth_gdf.iterrows():
        true_poly = kwimage.MultiPolygon.coerce(row['geometry'])
        truth_colorized = true_poly.draw_on(truth_colorized, color=row['color'])
        if row['status'] not in {'negative', 'ignore'}:
            truth_labels = true_poly.fill(truth_labels, value=idx)
            idx += 1

    truth_labels
    start_time = heatmap_time_intervals[0][0]
    end_time = heatmap_time_intervals[-1][1]
    print('start_time = {}'.format(ub.urepr(start_time, nl=1)))
    print('end_time = {}'.format(ub.urepr(end_time, nl=1)))

    classes = data['classes']
    channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                      for cat in classes.cats.values()]

    unique_labels = np.unique(truth_labels)
    unique_colors = kwimage.Color.distinct(len(unique_labels))
    label_to_color = ub.dzip(unique_labels, unique_colors)
    label_to_color[0] = kwimage.Color.coerce('black').as01()

    truth_canvas = util_kwimage.colorize_label_image(truth_labels, label_to_color=label_to_color)

    frame0_canvas = kwarray.robust_normalize(heatmap_thwc[0])
    frame0_canvas = util_kwimage.perchannel_colorize(frame0_canvas, channel_colors=channel_colors)
    frame0_canvas = util_kwimage.ensure_false_color(frame0_canvas)
    frame0_canvas = kwimage.nodata_checkerboard(frame0_canvas)
    frame0_canvas = bounds.draw_on(frame0_canvas, fill=False, edgecolor='white')

    frameN_canvas = kwarray.robust_normalize(heatmap_thwc[-1])
    frameN_canvas = util_kwimage.perchannel_colorize(frameN_canvas, channel_colors=channel_colors)
    frameN_canvas = kwimage.nodata_checkerboard(frameN_canvas)
    frameN_canvas = bounds.draw_on(frameN_canvas, fill=False, edgecolor='white')

    truth_canvas = bounds.draw_on(truth_canvas, fill=False, edgecolor='white')

    kwplot.autompl()
    kwplot.imshow(frame0_canvas, title=f'Frame 0 Colorized: {start_time}', fnum=1)
    kwplot.imshow(frameN_canvas, title=f'Frame N Colorized: {end_time}', fnum=2)
    kwplot.imshow(truth_canvas, title='Truth Labels', fnum=3)
    kwplot.imshow(truth_colorized, title='Truth Status', fnum=4)

    self = PolygonExtractor(heatmap_thwc, bounds=bounds,
                            heatmap_time_intervals=heatmap_time_intervals,
                            classes=classes, config={
                                # 'algo': 'meanshift',
                                # 'algo': 'crall',
                                # 'algo': 'leotta',
                            })

    self.config['algo'] = 'crall'
    self.config['scale_factor'] = 1

    polygons, info = self.predict_polygons(return_info=True)
    label_mask = info['label_mask']

    canvas = label_mask.colorize()
    for p in polygons:
        canvas = p.draw_on(canvas, fill=0, edgecolor='black')
    kwplot.imshow(canvas, pnum=(1, 2, 2), fnum=6)

    stacked = self.draw_timesequence()
    kwplot.imshow(stacked, pnum=(1, 2, 1), fnum=6)


def real_data_demo_case_1_fixed():
    from geowatch.utils import util_girder
    from geowatch.utils import util_kwimage
    import pickle
    import kwimage
    import kwarray
    import kwplot
    import numpy as np

    api_url = 'https://data.kitware.com/api/v1'
    data_fpath = ub.Path(util_girder.grabdata_girder(
        api_url, '654eab24314693d6b1df2658', hash_prefix='358bc1e7ec9ab7ba'))

    data = pickle.loads(data_fpath.read_bytes())
    heatmap_thwc = data['heatmap_thwc']
    bounds = kwimage.MultiPolygon.from_shapely(data['bas_gdf'].geometry.unary_union)
    heatmap_time_intervals = data['heatmap_time_intervals']
    truth_gdf = data['truth_gdf']

    t, h, w, c = heatmap_thwc.shape

    truth_colorized = np.ones((h, w, 3))
    truth_labels = np.zeros((h, w), dtype=np.int32)
    idx = 1
    for _, row in truth_gdf.iterrows():
        true_poly = kwimage.MultiPolygon.coerce(row['geometry'])
        truth_colorized = true_poly.draw_on(truth_colorized, color=row['color'])
        if row['status'] not in {'negative', 'ignore'}:
            truth_labels = true_poly.fill(truth_labels, value=idx)
            idx += 1

    truth_labels
    start_time = heatmap_time_intervals[0][0]
    end_time = heatmap_time_intervals[-1][1]
    print('start_time = {}'.format(ub.urepr(start_time, nl=1)))
    print('end_time = {}'.format(ub.urepr(end_time, nl=1)))

    classes = data['classes']
    channel_colors = [kwimage.Color.coerce(cat['color']).as01()
                      for cat in classes.cats.values()]

    unique_labels = np.unique(truth_labels)
    unique_colors = kwimage.Color.distinct(len(unique_labels))
    label_to_color = ub.dzip(unique_labels, unique_colors)
    label_to_color[0] = kwimage.Color.coerce('black').as01()

    truth_canvas = util_kwimage.colorize_label_image(truth_labels, label_to_color=label_to_color)

    frame0_canvas = kwarray.robust_normalize(heatmap_thwc[0])
    frame0_canvas = util_kwimage.perchannel_colorize(frame0_canvas, channel_colors=channel_colors)
    frame0_canvas = util_kwimage.ensure_false_color(frame0_canvas)
    frame0_canvas = kwimage.nodata_checkerboard(frame0_canvas)
    frame0_canvas = bounds.draw_on(frame0_canvas, fill=False, edgecolor='white')

    frameN_canvas = kwarray.robust_normalize(heatmap_thwc[-1])
    frameN_canvas = util_kwimage.perchannel_colorize(frameN_canvas, channel_colors=channel_colors)
    frameN_canvas = kwimage.nodata_checkerboard(frameN_canvas)
    frameN_canvas = bounds.draw_on(frameN_canvas, fill=False, edgecolor='white')

    truth_canvas = bounds.draw_on(truth_canvas, fill=False, edgecolor='white')

    kwplot.autompl()
    kwplot.imshow(frame0_canvas, title=f'Frame 0 Colorized: {start_time}', fnum=1)
    kwplot.imshow(frameN_canvas, title=f'Frame N Colorized: {end_time}', fnum=2)
    kwplot.imshow(truth_canvas, title='Truth Labels', fnum=3)
    kwplot.imshow(truth_colorized, title='Truth Status', fnum=4)

    self = PolygonExtractor(heatmap_thwc, bounds=bounds,
                            heatmap_time_intervals=heatmap_time_intervals,
                            classes=classes, config={})

    # self.config['algo'] = 'meanshift'
    # self.config['algo'] = 'leotta'
    # self.config['leotta_threah'] = 0.2
    # self.config['algo'] = 'kmeans'
    self.config['algo'] = 'crall'
    self.config['scale_factor'] = 8

    polygons, info = self.predict_polygons(return_info=True)
    label_mask = info['label_mask']

    canvas = label_mask.resize(scale=4).colorize()
    for p in polygons:
        canvas = p.scale(4).draw_on(canvas, fill=0, edgecolor='black')
    kwplot.imshow(canvas, pnum=(1, 2, 1), fnum=6)

    stacked = self.draw_timesequence()
    kwplot.imshow(stacked, pnum=(1, 2, 1), fnum=6)


def generate_real_example():
    import kwcoco
    import kwutil
    import pickle  # NOQA
    import numpy as np
    import kwimage  # NOQA
    from geowatch.cli import reproject_annotations

    # Read BAS dataset
    dset = kwcoco.CocoDataset('/home/joncrall/data/dvc-repos/smart_expt_dvc/_airflow/preeval17_batch_v130/KR_R002/sc-fusion/sc_fusion_kwcoco_tracked.json')

    # Classes of interest
    classes = kwcoco.CategoryTree.from_coco([
        {'id': 30, 'name': 'No Activity', 'color': 'tomato'},
        {'id': 31, 'name': 'Site Preparation', 'color': 'gold'},
        {'id': 32, 'name': 'Active Construction', 'color': 'lime'},
        {'id': 33, 'name': 'Post Construction', 'color': 'darkturquoise'},
        {'id': 34, 'name': 'ac_salient', 'color': 'white'},
    ])

    # from geowatch.geoannots import geomodels
    # region = geomodels.RegionModel.coerce('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/annotations/drop7/region_models/KR_R002.geojson')

    # Create a copy with the Truth
    true_dset = dset.copy()
    sites = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/annotations/drop7/site_models/KR_R002*.geojson'
    true_dset = reproject_annotations.main(dst='return', src=true_dset, sites=sites, role='truth')

    # Choose a Video
    video_id = dset.videos()[0]
    images = dset.images(video_id=video_id)

    resolution = '8GSD'

    # Construct spatial truth and prediction summary
    bas_gdf = coco_make_track_gdf(dset, video_id, resolution)
    truth_gdf = coco_make_track_gdf(true_dset, video_id, resolution)
    bounds = kwimage.MultiPolygon.from_shapely(bas_gdf.geometry.unary_union)

    # Figure out the shape of the canvas at the desired resolution
    delayed = images.coco_images[0].imdelay(channels='red', space='video', resolution=resolution)
    h, w = delayed.shape[0:2]

    import kwplot
    positive_true_polys = []
    nonpositive_true_polys = []
    status_to_color = {}
    for _, row in truth_gdf.iterrows():
        true_poly = kwimage.MultiPolygon.coerce(row['geometry'])
        status_to_color[row['status']] = row['color']
        if row['status'] != {'ignore', 'negative'}:
            positive_true_polys.append(true_poly)

    status_to_color['BAS output'] = 'kitware_blue'

    if 1:
        import kwarray
        coco_img = images.coco_images[0]
        rgb = coco_img.imdelay(channels='red|green|blue', space='video', resolution=resolution).finalize()
        # dets = coco_img._detections_for_resolution(space='video', resolution=resolution)
        canvas = kwarray.robust_normalize(rgb)
        # true_coco_image = true_dset.coco_image(coco_img['id'])
        # true_dets = true_coco_image._detections_for_resolution(space='video')
        positive_true_dets = kwimage.Detections(segmentations=positive_true_polys)
        nonpositive_true_dets = kwimage.Detections(segmentations=nonpositive_true_polys)
        canvas = positive_true_dets.data['segmentations'].draw_on(canvas, alpha=0.5, color='kitware_green', edgecolor='black')
        canvas = nonpositive_true_dets.data['segmentations'].draw_on(canvas, alpha=0.5, color='kitware_gray', edgecolor='black')

        canvas = bounds.draw_on(canvas, color='kitware_blue', alpha=0.5)
        kwplot.imshow(canvas, fnum=2)

    channels = 'Site Preparation|Active Construction|Post Construction|No Activity|ac_salient'
    print(f'Reading heatmaps with: key={channels} @ {resolution}')
    io_workers = kwutil.util_parallel.coerce_num_workers('avail')
    heatmap_iter = imread_many(dset, images, channels=channels, space='video', resolution=resolution, workers=io_workers)
    _heatmaps = list(heatmap_iter)
    heatmap_thwc = np.array(_heatmaps)

    image_dates = [kwutil.util_time.coerce_datetime(d)
                   for d in images.lookup('date_captured')]
    heatmap_time_intervals = list(zip(image_dates, image_dates))

    data = {}
    data['heatmap_thwc'] = heatmap_thwc
    data['heatmap_time_intervals'] = heatmap_time_intervals
    data['bounds'] = bounds
    data['classes'] = classes
    data['bas_gdf'] = bas_gdf
    data['truth_gdf'] = truth_gdf

    if 1:
        # Splot check
        kwargs = ub.compatible(data, PolygonExtractor.__init__)
        self = PolygonExtractor(**kwargs)
        import kwplot
        kwplot.autompl()
        kwplot.imshow(self.draw_timesequence())

    if 0:
        # Upload data
        pickle_fpath = ub.Path.appdir('geowatch/polyextract').ensuredir() / 'demo0-fixed.pkl'
        pickle_fpath.write_bytes(pickle.dumps(data))

        hash_prefix = ub.hash_file(pickle_fpath)[0:16]
        print(hash_prefix)

        # source $HOME/internal/secrets
        ub.cmd(ub.paragraph(
            f'''
            . ~/internal/safe/secrets && \
            girder-client \
                --api-url https://data.kitware.com/api/v1 \
                upload \
                654d10a9fdc1508d8bec1bb6 \
                {pickle_fpath}
            '''), verbose=3, shell=True)


class LabelMask:
    """
    Ignore:
        from geowatch.utils import util_kwimage
        import kwplot
        kwplot.autompl()
        self = LabelMask.demo()
        canvas = self.colorize()
        kwplot.imshow(canvas, doclf=1)
        polys = self.to_multi_polygons()
        for p in polys:
            p.draw(fill=0, edgecolor='black')
    """
    def __init__(self, data):
        self.data = data

    def colorize(self):
        from geowatch.utils import util_kwimage
        canvas = util_kwimage.colorize_label_image(self.data, legend_dpi=300)
        return canvas

    def resize(self, scale=None, dsize=None):
        import kwimage
        new_data = kwimage.imresize(self.data, scale=scale, interpolation='nearest')
        new = self.__class__(new_data)
        return new

    @classmethod
    def demo(LabelMask):
        import numpy as np
        import kwimage
        H, W = 512, 512
        label_img = np.zeros((H, W), dtype=np.int32)
        poly1 = kwimage.Polygon.random().scale((H, W))
        poly2 = kwimage.Polygon.random().scale((H, W))
        label_img = poly1.fill(label_img, value=1)
        label_img = poly2.fill(label_img, value=2)
        self = LabelMask(label_img)
        return self

    def to_multi_polygons(self):
        import kwimage
        label_img = self.data
        poly_datas = _rasterio_labelimg_to_polygons(label_img)
        value_to_polys = ub.group_items(poly_datas, key=lambda p: p['value'])
        multi_polygons = []
        for value, subpolys in value_to_polys.items():
            kwpolys = [
                kwimage.Polygon(**p, metakeys=['value'])
                for p in subpolys
            ]
            mpoly = kwimage.MultiPolygon(kwpolys, meta={'value': value})
            multi_polygons.append(mpoly)
        return multi_polygons


def _rasterio_labelimg_to_polygons(label_img):
    """
    Note:
        The :func:`rasterio.features.shapes` is capable of multi-label polygon
        extraction.

    TODO:
        - [ ] Kwimage needs a good multi-label image polygon extraction tool

    Ignore:
    """
    import numpy as np
    from rasterio import features
    poly_datas = []
    if label_img.size > 0:
        shapes = list(features.shapes(label_img, connectivity=8))
        translate = np.array([-0.5, -0.5]).ravel()[None, :]
        for shape, value in shapes:
            if value > 0:
                coords = shape['coordinates']
                exterior = np.array(coords[0]) + translate
                interiors = [np.array(p) + translate for p in coords[1:]]
                poly_datas.append({
                    'exterior': exterior,
                    'interiors': interiors,
                    'value': value,
                })
    return poly_datas


def coco_make_track_gdf(coco_dset, video_id, resolution=None):
    import geopandas as gpd
    from shapely.ops import unary_union
    from geowatch import heuristics
    coco_images = coco_dset.images(video_id=video_id).coco_images
    tid_to_objs = ub.ddict(list)
    for coco_img in coco_images:
        annots = coco_img.annots()
        track_ids = annots.lookup('track_id')
        polys = coco_img._annot_segmentations(annots.objs, space='video', resolution=resolution)
        ann0 = annots.objs[0]
        status = ann0.get('status', None)
        for tid, poly, ann in zip(track_ids, polys, annots.objs):
            ann['poly'] = poly.to_shapely()
            tid_to_objs[tid].append(ann)

    summary_info = []
    for tid, objs in tid_to_objs.items():
        obj0 = objs[0]
        status = obj0.get('status', None)
        track_poly = unary_union([obj['poly'] for obj in objs])
        if status is None:
            color = 'kitware_blue'
        else:
            color = heuristics.IARPA_STATUS_TO_INFO[status]['color']
        summary_info.append({
            'track_id': tid,
            'status': status,
            'color': color,
            'geometry': track_poly,
        })
    gdf = gpd.GeoDataFrame(summary_info)
    return gdf


def imread_many(dset, gids, channels=None, space='video', resolution=None, workers=0):
    import kwutil
    load_jobs = ub.JobPool(mode='process', max_workers=workers)
    pman = kwutil.util_progress.ProgressManager()
    with load_jobs, pman:
        for gid in pman.progiter(gids, desc=f'submit load images jobs: {channels}'):
            coco_img = dset.coco_image(gid)
            delayed = coco_img.imdelay(channels=channels, space='video', resolution=resolution)
            load_jobs.submit(delayed.finalize)

        for job in pman.progiter(load_jobs.jobs, desc=f'collect load images jobs: {channels}'):
            imdata = job.result()
            yield imdata



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Axial-Blocks proposed in Axial-DeepLab [1].

Axial-Blocks are based on residual bottleneck blocks, but with the 3x3
convolution replaced with two axial-attention layers, one on the height-axis,
followed by the other on the width-axis.

[1] Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
    ECCV 2020 Spotlight.
      Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
      Liang-Chieh Chen.
"""
import tensorflow as tf

from deeplab2.model import utils
from deeplab2.model.layers import activations
from deeplab2.model.layers import axial_layers
from deeplab2.model.layers import convolutions
from deeplab2.model.layers import squeeze_and_excite


class AxialBlock(tf.keras.layers.Layer):
    """An AxialBlock as a building block for an Axial-ResNet model.

    We implement the Axial-Block proposed in [1] in a general way that also
    includes convolutional residual blocks, such as the basic block and the
    bottleneck block (w/ and w/o Switchable Atrous Convolution).

    A basic block consists of two 3x3 convolutions and a residual connection. It
    is the main building block for wide-resnet variants.

    A bottleneck block consists of consecutive 1x1, 3x3, 1x1 convolutions and a
    residual connection. It is the main building block for standard resnet
    variants.

    An axial block consists of a 1x1 input convolution, a self-attention layer
    (either axial-attention or global attention), a 1x1 output convolution, and a
    residual connection. It is the main building block for axial-resnet variants.

    Note: We apply the striding in the first spatial operation (i.e. 3x3
    convolution or self-attention layer).
    """

    def __init__(self,
                 filters_list,
                 kernel_size=3,
                 strides=1,
                 atrous_rate=1,
                 use_squeeze_and_excite=False,
                 use_sac=False,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu',
                 name=None,
                 conv_kernel_weight_decay=0.0,
                 basic_block_second_conv_atrous_rate=None,
                 attention_type=None,
                 axial_layer_config=None):
        """Initializes an AxialBlock.

        Args:
          filters_list: A list of filter numbers in the residual block. We currently
            support filters_list with two or three elements. Two elements specify
            the filters for two consecutive 3x3 convolutions, while three elements
            specify the filters for three convolutions (1x1, 3x3, and 1x1).
          kernel_size: The size of the convolution kernels (default: 3).
          strides: The strides of the block (default: 1).
          atrous_rate: The atrous rate of the 3x3 convolutions (default: 1). If this
            residual block is a basic block, it is recommendeded to specify correct
            basic_block_second_conv_atrous_rate for the second 3x3 convolution.
            Otherwise, the second conv will also use atrous rate, which might cause
            atrous inconsistency with different output strides, as tested in
            axial_block_groups_test.test_atrous_consistency_basic_block.
          use_squeeze_and_excite: A boolean flag indicating whether
            squeeze-and-excite (SE) is used.
          use_sac: A boolean, using the Switchable Atrous Convolution (SAC) or not.
          bn_layer: A tf.keras.layers.Layer that computes the normalization
            (default: tf.keras.layers.BatchNormalization).
          activation: A string specifying the activation function to apply.
          name: An string specifying the name of the layer (default: None).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.
          basic_block_second_conv_atrous_rate: An integer, the atrous rate for the
            second convolution of basic block. This is necessary to ensure atrous
            consistency with different output_strides. Defaults to atrous_rate.
          attention_type: A string, type of attention to apply. Support 'axial' and
            'global'.
          axial_layer_config: A dict, an argument dictionary for the axial layer.

        Raises:
          ValueError: If filters_list does not have two or three elements.
          ValueError: If attention_type is not supported.
          ValueError: If double_global_attention is True in axial_layer_config.
        """
        super(AxialBlock, self).__init__(name=name)

        self._filters_list = filters_list
        self._strides = strides
        self._use_squeeze_and_excite = use_squeeze_and_excite
        self._bn_layer = bn_layer
        self._activate_fn = activations.get_activation(activation)
        self._attention_type = attention_type

        if axial_layer_config is None:
            axial_layer_config = {}

        if basic_block_second_conv_atrous_rate is None:
            basic_block_second_conv_atrous_rate = atrous_rate

        if len(filters_list) == 3:
            # Three consecutive convolutions: 1x1, 3x3, and 1x1.
            self._conv1_bn_act = convolutions.Conv2DSame(
                filters_list[0], 1, 'conv1_bn_act',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation=activation,
                conv_kernel_weight_decay=conv_kernel_weight_decay)

            if attention_type is None or attention_type.lower() == 'none':
                self._conv2_bn_act = convolutions.Conv2DSame(
                    filters_list[1], kernel_size, 'conv2_bn_act',
                    strides=strides,
                    atrous_rate=atrous_rate,
                    use_bias=False,
                    use_bn=True,
                    bn_layer=bn_layer,
                    activation=activation,
                    use_switchable_atrous_conv=use_sac,
                    # We default to use global context in SAC if use_sac is True. This
                    # setting is experimentally found effective.
                    use_global_context_in_sac=use_sac,
                    conv_kernel_weight_decay=conv_kernel_weight_decay)
            elif attention_type == 'axial':
                if 'double_global_attention' in axial_layer_config:
                    if axial_layer_config['double_global_attention']:
                        raise ValueError('Double_global_attention takes no effect in '
                                         'AxialAttention2D.')
                    del axial_layer_config['double_global_attention']
                self._attention = axial_layers.AxialAttention2D(
                    strides=strides,
                    filters=filters_list[1],
                    name='attention',
                    bn_layer=bn_layer,
                    conv_kernel_weight_decay=conv_kernel_weight_decay,
                    **axial_layer_config)
            elif attention_type == 'global':
                self._attention = axial_layers.GlobalAttention2D(
                    strides=strides,
                    filters=filters_list[1],
                    name='attention',
                    bn_layer=bn_layer,
                    conv_kernel_weight_decay=conv_kernel_weight_decay,
                    **axial_layer_config)
            else:
                raise ValueError(attention_type + ' is not supported.')

            # Here we apply a batch norm with gamma initialized at zero. This ensures
            # that at random initialization of the model, the skip connections
            # dominate all residual blocks. In this way, all the skip connections
            # construct an identity mapping that passes the gradients (without any
            # distortion from the randomly initialized blocks) to all residual blocks.
            # This trick helps training at early epochs.
            # Reference: "Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour".
            # https://arxiv.org/abs/1706.02677
            self._conv3_bn = convolutions.Conv2DSame(
                filters_list[2], 1, 'conv3_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay)
        elif len(filters_list) == 2:
            # Two consecutive convolutions: 3x3 and 3x3.
            self._conv1_bn_act = convolutions.Conv2DSame(
                filters_list[0], kernel_size, 'conv1_bn_act',
                strides=strides,
                atrous_rate=atrous_rate,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation=activation,
                use_switchable_atrous_conv=use_sac,
                use_global_context_in_sac=use_sac,
                conv_kernel_weight_decay=conv_kernel_weight_decay)
            # Here we apply a batch norm with gamma initialized at zero. This ensures
            # that at random initialization of the model, the skip connections
            # dominate all residual blocks. In this way, all the skip connections
            # construct an identity mapping that passes the gradients (without any
            # distortion from the randomly initialized blocks) to all residual blocks.
            # This trick helps training at early epochs.
            # Reference: "Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour".
            # https://arxiv.org/abs/1706.02677
            self._conv2_bn = convolutions.Conv2DSame(
                filters_list[1], kernel_size, 'conv2_bn',
                strides=1,
                atrous_rate=basic_block_second_conv_atrous_rate,
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                use_switchable_atrous_conv=use_sac,
                use_global_context_in_sac=use_sac,
                conv_kernel_weight_decay=conv_kernel_weight_decay)
        else:
            raise ValueError('Expect filters_list to have length 2 or 3; got %d' %
                             len(filters_list))

        if self._use_squeeze_and_excite:
            self._squeeze_and_excite = squeeze_and_excite.SimplifiedSqueezeAndExcite(
                filters_list[-1])
        self._conv_kernel_weight_decay = conv_kernel_weight_decay

    def build(self, input_shape_list):
        input_tensor_shape = input_shape_list[0]
        self._shortcut = None
        if input_tensor_shape[3] != self._filters_list[-1]:
            self._shortcut = convolutions.Conv2DSame(
                self._filters_list[-1], 1, 'shortcut',
                strides=self._strides,
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                activation='none',
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

    def call(self, inputs):
        """Performs a forward pass.

        We have to define drop_path_random_mask outside the layer call and pass it
        into the layer, because recompute_grad (gradient checkpointing) does not
        allow any randomness within the function call. In addition, recompute_grad
        only supports float tensors as inputs. For this reason, the training flag
        should be also passed as a float tensor. For the same reason, we cannot
        support passing drop_path_random_mask as None. Instead, we ask the users to
        pass only the first two tensors when drop path is not used.

        Args:
          inputs: A tuple of 2 or 3 tensors, containing
            input_tensor should be an input tensor of type tf.Tensor with shape
              [batch, height, width, channels].
            float_tensor_training should be a float tensor of 0.0 or 1.0, whether
              the model is in training mode.
            (optional) drop_path_random_mask is a drop path random mask of type
              tf.Tensor with shape [batch, 1, 1, 1].

        Returns:
          outputs: A single tensor without activation function.

        Raises:
          ValueError: If the length of inputs is not 2 or 3.
        """
        if len(inputs) not in (2, 3):
            raise ValueError('The length of inputs should be either 2 or 3.')

        # Unpack the inputs.
        input_tensor, float_tensor_training, drop_path_random_mask = (
            utils.pad_sequence_with_none(inputs, target_length=3))

        # AxialBlock expects an activated feature as input_tensor.
        input_tensor = self._activate_fn(input_tensor)

        # Recompute_grad takes only float tensors as inputs. It does not allow
        # bools or boolean tensors. For this reason, we cast training to a float
        # tensor outside this call, and now we cast it back to a boolean tensor.
        training = tf.cast(float_tensor_training, tf.bool)

        shortcut = input_tensor
        if self._shortcut is not None:
            shortcut = self._shortcut(shortcut, training=training)
        elif self._strides != 1:
            shortcut = shortcut[:, ::self._strides, ::self._strides, :]

        if len(self._filters_list) == 3:
            x = self._conv1_bn_act(input_tensor, training=training)
            if (self._attention_type is None or
                    self._attention_type.lower() == 'none'):
                x = self._conv2_bn_act(x, training=training)
            else:
                x = self._attention(x, training=training)
                x = self._activate_fn(x)
            x = self._conv3_bn(x, training=training)
        if len(self._filters_list) == 2:
            x = self._conv1_bn_act(input_tensor, training=training)
            x = self._conv2_bn(x, training=training)

        if self._use_squeeze_and_excite:
            x = self._squeeze_and_excite(x)

        if drop_path_random_mask is not None:
            x = x * drop_path_random_mask
        x = x + shortcut
        return x
